/*
** The spreadsheet (ss) widget
*/
static char const rcsid[] = "@(#) $Header: /home/softools/cvs/softools/code/common/cked/src/ss.cpp,v 1.88 2004/04/05 19:32:40 mark Exp $";

#include <tcl.h>
#include <tk.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include "ss.h"

#if INTERFACE

/*
** object: ss
**
** A spreadsheet is represented in memory by a single object of the
** ss class.
*/
class ss {
private:
  Tcl_Interp *interp;      // The Tcl interpreter
  char *zSs;               // The name of the spreadsheet widget.  Malloc'ed.
  SsCanvas fastcan;        // The underlying "fast" Tk canvas.
  SsConfigure sConfigure;  // The configurable options for the ss widget

  int mouseMoved;          // Nonzero if the mouse has moved since beginning
                           // the selection

  SsCell selectionAnchor;  // The cell that is the current selection anchor
                           // (for the spreadsheet-level selection)
  SsCell activeCell;       // The cell that is currently active; that is,
                           // contains the insertion cursor.
  int fromSs;              // Nonzero if a traced variable modification was 
                           // initiated internally by the widget
  int nDisplayedRows;      // The number of spreadsheet rows currently drawn
  SsCell topleft;          // The non-title cell currently being displayed at
                           // the top left
  int displayProperties;   // An OR-ed list of properties managing the
                           // redisplay of the spreadsheet.
  int cellPad;             // Number of pixels to internally pad each cell.
  int currentTotalWidth;   // The width of all columns added together.
  int bDirtyCell;          // TRUE if at least one cell value changed.
                           // Used while redisplaying.
  int lastRowWithData;     // Cached value of the last row in the spreadsheet
                           // that contains displayable data.  Valid only if
                           // bDirtyCell is 0.
  Tk_Font font;            // The currently configured text font.

  int currentTitleRowHeight;  // The current height of title row.
                              // Needed if the default title row height is
                              // too small to contain its text.

  SsTclList *pSelectionRect; // A list describing the selection rectangle.
                             // The list consists of the top row, left column,
                             // bottom row, and right column around which the
                             // rectangle is currently drawn.

  SsArrayOfPointer aHeight;   // Array mapping rows to their heights
  SsArrayOfPointer aWidth;    // Array mapping columns to their widths
  SsArrayOfPointer aHidden;   // Array mapping columns to their hidden state
  SsArrayOfPointer aMinWidth; // Array mapping columns to their minimum widths
  SsStringArrayOfString aBg;  // Array mapping cells to background colors
  SsStringArrayOfString aFg;  // Array mapping cells to foreground colors
};

/*
** Bits for ss::displayProperties.
*/
#define DISPLAY_PENDING         0x00000001  // Redisplay the spreadsheet
#define DISPLAY_TRIM_ROWS       0x00000002  // Trim the number of rows
                                            // in the spreadsheet, if possible
#define DISPLAY_ADD_ROWS        0x00000004  // Add to the number of rows,
                                            // if possible
/*
** object: SsConfigure
**
** The configuration for a single ss widget is contained in an SsConfigure
** object.  Note that all "char *" members contain
** either 0 or an allocated string, except for zBdColor.
*/
class SsConfigure {
private:
  SsArrayOfString aOptions; // Hash table that links the option name to
                            // its index within the option spec table
public:
  char *zBg;            // The default background color of all cells
  char *zFg;            // The default foreground (text) color of all cells
  char *zActiveBg;      // The background color of the active cell
  char *zActiveFg;      // The foreground color of the active cell
  char *zSelectBg;      // The background color of a selected cell
  char *zSelectFg;      // The foreground (text) color of a selected cell
  char *zTitleBg;       // The background color of a title cell
  char *zTitleFg;       // The foreground (text) color of a title cell
  int cols;             // The default number of columns in the spreadsheet
  int rows;             // The default number of rows in the spreadsheet
  int colWidth;         // The default width for any column, in pixels
  int rowHeight;        // The default height of any row, in pixels
  char *zCursor;        // The cursor to draw in a cell
  char *zFont;          // The font for text
  int height;           // The desired height of the spreadsheet, in rows
  int width;            // The desired width of the spreadsheet, in columns
  int maxRows;          // The maximum number of rows that can appear
  int minColWidth;      // The minimum width of any one column, in pixels
  int minRowHeight;     // The minimum height of any row, in pixels
  int expandCol;        // The column that will be automatically expanded when
                        // the screen dimensions change.  -1 if no expansion
                        // is to be done.
  char *zResizeColor;   // The color of a line used for a row or column resize
  char *zHResizeCursor; // The cursor when the pointer is over a horizontal
                        // resize line
  char *zVResizeCursor; // The cursor when the pointer is over a vertical
                        // resize line
  char *zTitleCursor;   // The name of the cursor displayed when the mouse is
                        // over a title cell
  char *zState;         // The state of the spreadsheet: "normal" or "disabled"
  char *zIfDisabled;    // This value qualifies the value in zState, if
                        // zState=="disabled".  If zIfDisabled is "all"
                        // (the default), then all interaction with the
                        // spreadsheet is disabled.  But if its value is
                        // "table", then only table changes are disabled--
                        // the user can still change the row heights, etc.
  int titleRowHeight;   // The height of a title cell in the header row
  int titleColWidth;    // The width of a title cell in the header column
  char *zVariable;      // The name of the widget Tcl variable, which will 
                        // contain any entry for each cell value,
                        // as well as other spreadsheet information
  char *zXScrollCmd;    // The Tcl command used to scroll horizontally
  char *zYScrollCmd;    // The Tcl command used to scroll vertically
  char *zBdColor;       // Border color for the spreadsheet.
  char *zWaterImage;    // Name of the image used for watermark.  0 if none.
  char *zHideColor;     // Color of vertical line representing hidden columns.
                        // "gray50" is the default.
}

/*
** object: SsCell
**
** A cell in the spreadsheet is represented by a single object
** of the SsCell class.
*/
class SsCell {
private:
  char zCell[2*32+2];  // String rep.  Large enough to fit 2 32-bit integers,
                       // plus a comma.
  int dirty;           // Nonzero if zCell isn't valid
  int row;
  int col;
public:
  SsCell::SsCell(const SsCell &cell) {row=cell.row; col=cell.col; dirty=1;}
  SsCell::SsCell(const SsCell *pCell) {set(pCell);}
  SsCell::SsCell(const char *zCell) {set(zCell);}
  SsCell::SsCell() {row=-1; col=-1; dirty=1;}
  SsCell::SsCell(int nRow, int nCol) {row=nRow; col=nCol; dirty=1;}

  int isnull() {return (row==-1 && col==-1);}
  char *render() {
    if( dirty ){
      if( isnull() ) {
        zCell[0] = 0;
      } else {
        sprintf(zCell, "%d,%d", row, col);
      }
      dirty = 0;
    }
    return zCell;
  }
  void set(const char *zNewCell){
    if( !zNewCell || sscanf(zNewCell, "%d,%d", &row, &col)!=2 ){
      row=-1; col=-1;
    }
    dirty = 1;
  }
  void set(const SsCell *pCell){
    if( pCell ){
      row=pCell->row; col=pCell->col;
    } else {
      row=-1; col=-1;
    }
    dirty = 1;
  }
  void set(const SsCell &cell) {set(&cell);}
  void set(int nRow, int nCol) {row = nRow; col = nCol; dirty = 1;}
  void clear(){row = -1; col = -1; dirty = 1;}
  void SetRow(int nRow) {row = nRow; dirty = 1;}
  void SetCol(int nCol) {col = nCol; dirty = 1;}
  int GetRow() {return row;}
  int GetCol() {return col;}
  int equal(const SsCell &cell) {return (cell.row==row && cell.col==col);}
  int equal(const SsCell *pCell) {
    return (pCell && pCell->row==row && pCell->col==col);
  }
}

/*
** object: SsBbox
**
** A bounding box is represented by a single object of the 
** SsBbox class.
*/
class SsBbox {
private:
  char zBbox[32*4+4];  // large enough to always fit 4 integers, 3 spaces, 
                       // and a null
  int dirty;
public:
  int x0, x1;
  int y0, y1;
  SsBbox::SsBbox(const char *zNewBbox) {
    if( sscanf(zNewBbox, "%d %d %d %d", &x0, &y0, &x1, &y1)!=4 ){
      x0 = x1 = y0 = y1 = -1;
    }
    zBbox[0] = 0;
    dirty = 1;
  }
  char *render() {
    if( dirty ) {
      sprintf(zBbox, "%d %d %d %d", x0, y0, x1, y1); 
      dirty = 0;
    }
    return zBbox;
  }
  int GetWidth() {return abs(x1 - x0) + 1;}
  int GetHeight() {return abs(y1 - y0) + 1;}
  int isnull() {return (x0==-1 && x1==-1 && y0==-1 && y1==-1);}
}
#endif

/*
** The following structure contains the Tk configuration options for the
** ss widget
*/
static SsConfigure off;
#define My_Offset(obj, member) ((int) ((char *) &(obj.member) - (char*)(&obj)))
static const Tk_ConfigSpec ssSpecs[] = {

  {TK_CONFIG_STRING, "-activebackground", "activeBackground",
     "ActiveBackground", "lightblue", My_Offset(off, zActiveBg), 0},
  {TK_CONFIG_SYNONYM, "-activebg", "activeBackground", (char *)NULL, 
      (char *)NULL, 0, 0},

  {TK_CONFIG_STRING, "-activeforeground", "activeForeground",
     "ActiveForeground", "", My_Offset(off, zActiveFg), 0},
  {TK_CONFIG_SYNONYM, "-activefg", "activeForeground", (char *)NULL, 
      (char *)NULL, 0, 0},

  {TK_CONFIG_STRING, "-background", "background",
     "Background", "white", My_Offset(off, zBg), 0},
  {TK_CONFIG_SYNONYM, "-bg", "background", (char *)NULL, (char *)NULL, 0, 0},

  {TK_CONFIG_INT, "-cols", "cols",
     "Cols", "1", My_Offset(off, cols), 0},

  {TK_CONFIG_INT, "-colwidth", "colWidth",
     "ColWidth", "60", My_Offset(off, colWidth), 0},

  {TK_CONFIG_STRING, "-cursor", "cursor",
     "Cursor", "xterm", My_Offset(off, zCursor), 0},

  {TK_CONFIG_INT, "-expandcol", "expandCol",
     "ExpandCol", "-1", My_Offset(off, expandCol), 0},

  {TK_CONFIG_STRING, "-font", "font",
     "Font", "Helvetica 10", My_Offset(off, zFont), 0},

  {TK_CONFIG_STRING, "-foreground", "foreground",
     "Foreground", "black", My_Offset(off, zFg), 0},
  {TK_CONFIG_SYNONYM, "-fg", "foreground", (char *)NULL, (char *)NULL, 0, 0},

  {TK_CONFIG_INT, "-height", "height",
     "Height", "0", My_Offset(off, height), 0},

  {TK_CONFIG_STRING, "-hidecolor", "hidecolor",
     "HideColor", "gray50", My_Offset(off, zHideColor), 0},

  {TK_CONFIG_STRING, "-hresizecursor", "hResizeCursor",
     "HResizeCursor", "sb_v_double_arrow", 
     My_Offset(off, zHResizeCursor), 0},

  {TK_CONFIG_STRING, "-ifdisabled", "ifDisabled",
     "IfDisabled", "all", My_Offset(off, zIfDisabled), 0},

  {TK_CONFIG_INT, "-maxrows", "maxRows",
     "MaxRows", "500", My_Offset(off, maxRows), 0},

  {TK_CONFIG_INT, "-mincolwidth", "minColWidth",
     "MinColWidth", "12", My_Offset(off, minColWidth), 0},

  {TK_CONFIG_INT, "-minrowheight", "minRowHeight",
     "MinRowHeight", "12", My_Offset(off, minRowHeight), 0},

  {TK_CONFIG_STRING, "-resizecolor", "resizeColor",
     "ResizeColor", "white", My_Offset(off, zResizeColor), 0},

  {TK_CONFIG_INT, "-rowheight", "rowHeight",
     "RowHeight", "20", My_Offset(off, rowHeight), 0},

  {TK_CONFIG_INT, "-rows", "rows",
     "Rows", "1", My_Offset(off, rows), 0},

  {TK_CONFIG_STRING, "-selectbackground", "selectBackground",
     "SelectBackground", "red", My_Offset(off, zSelectBg), 0},
  {TK_CONFIG_SYNONYM, "-selectbg", "selectBackground", (char *)NULL, 
      (char *)NULL, 0, 0},

  {TK_CONFIG_STRING, "-selectforeground", "selectForeground",
     "SelectForeground", "", My_Offset(off, zSelectFg), 0},
  {TK_CONFIG_SYNONYM, "-selectfg", "selectForeground", (char *)NULL, 
      (char *)NULL, 0, 0},

  {TK_CONFIG_STRING, "-state", "state",
     "State", "normal", My_Offset(off, zState), 0},

  {TK_CONFIG_STRING, "-titlebackground", "titleBackground",
     "TitleBackground", "darkgray", My_Offset(off, zTitleBg), 0},
  {TK_CONFIG_SYNONYM, "-titlebg", "titleBackground", (char *)NULL, 
      (char *)NULL, 0, 0},

  {TK_CONFIG_INT, "-titlecolwidth", "titleColWidth",
     "TitleColWidth", "36", My_Offset(off, titleColWidth), 0},

  {TK_CONFIG_STRING, "-titlecursor", "titleCursor",
     "TitleCursor", "", My_Offset(off, zTitleCursor), 0},

  {TK_CONFIG_STRING, "-titleforeground", "titleForeground",
     "TitleForeground", "white", My_Offset(off, zTitleFg), 0},
  {TK_CONFIG_SYNONYM, "-titlefg", "titleForeground", (char *)NULL, 
      (char *)NULL, 0, 0},

  {TK_CONFIG_INT, "-titlerowheight", "titleRowHeight",
     "TitleRowHeight", "20", My_Offset(off, titleRowHeight), 0},

  {TK_CONFIG_STRING, "-variable", "variable",
     "Variable", "", My_Offset(off, zVariable), 0},

  {TK_CONFIG_STRING, "-vresizecursor", "vResizeCursor",
     "VResizeCursor", "sb_h_double_arrow", 
     My_Offset(off, zVResizeCursor), 0},

  {TK_CONFIG_STRING, "-watermark", "waterMark",
     "WaterMark", "", My_Offset(off, zWaterImage), 0},

  {TK_CONFIG_INT, "-width", "width",
     "Width", "0", My_Offset(off, width), 0},

  {TK_CONFIG_STRING, "-xscrollcommand", "xScrollCommand",
     "XScrollCommand", "", My_Offset(off, zXScrollCmd), 0},
  {TK_CONFIG_SYNONYM, "-xscrollcmd", "xScrollCommand", (char *)NULL, 
      (char *)NULL, 0, 0},

  {TK_CONFIG_STRING, "-yscrollcommand", "yScrollCommand",
     "YScrollCommand", "", My_Offset(off, zYScrollCmd), 0},
  {TK_CONFIG_SYNONYM, "-yscrollcmd", "yScrollCommand", (char *)NULL, 
      (char *)NULL, 0, 0},

  {TK_CONFIG_END, (char *)NULL, (char *)NULL, (char *)NULL,
     (char *)NULL, 0, 0},
};

/*
** Construct an ss object
*/
PUBLIC ss::ss(Tcl_Interp *pInterp, const char *zSsName){
  interp = pInterp;

  /* Create the canvas associated with the spreadsheet.
  */
  zSs = 0;
  ChangeStr(&zSs, zSsName);
  fastcan.init(pInterp, zSsName);

  /* Zero out all of the members
  */
  mouseMoved = 0;
  selectionAnchor.set(0, 0);
  activeCell.clear();
  fromSs = 0;
  nDisplayedRows = 0;
  topleft.set(1,1);

  displayProperties = 0;
  cellPad = 2;
  currentTotalWidth = 0;
  bDirtyCell = 1;
  lastRowWithData = 0;
  font = 0;
  currentTitleRowHeight = -1;
  pSelectionRect = new SsTclList(interp);

  sConfigure.zBg = 0;
  sConfigure.zFg = 0;
  sConfigure.zActiveBg = 0;
  sConfigure.zActiveFg = 0;
  sConfigure.zSelectBg = 0;
  sConfigure.zSelectFg = 0;
  sConfigure.zTitleBg = 0;
  sConfigure.zTitleFg = 0;
  sConfigure.cols = 0;
  sConfigure.colWidth = 0;
  sConfigure.rows = 0;
  sConfigure.rowHeight = 0;
  sConfigure.zCursor = 0;
  sConfigure.zFont = 0;
  sConfigure.height = 0;
  sConfigure.width = 0;
  sConfigure.maxRows = 0;
  sConfigure.minColWidth = 0;
  sConfigure.minRowHeight = 0;
  sConfigure.zResizeColor = 0;
  sConfigure.zHResizeCursor = 0;
  sConfigure.zVResizeCursor = 0;
  sConfigure.zTitleCursor = 0;
  sConfigure.zState = 0;
  sConfigure.zIfDisabled = 0;
  sConfigure.titleRowHeight = 0;
  sConfigure.titleColWidth = 0;
  sConfigure.zVariable = 0;
  sConfigure.zXScrollCmd = 0;
  sConfigure.zYScrollCmd = 0;
  sConfigure.zBdColor = "#a7a7a7";  // really ought to be calculated from
                                    // current value of sConfigure.zBg
  sConfigure.zHideColor = "gray50";
  sConfigure.zWaterImage = 0;
  sConfigure.expandCol = -1;

  /* Initialize the widget configuration from the defaults in the
  ** configuration spec
  */
  sConfigure.ConfigureWidgetDefaults();

  ResetCurrentWidth();
  Tk_Window tkwin = Tk_NameToWindow(interp, zSs, Tk_MainWindow(interp));
  if( sConfigure.zFont ){
    font = Tk_GetFont(interp, tkwin, sConfigure.zFont);
  }
}

/*
** Destroy a spreadsheet
*/
PUBLIC ss::~ss(){
  CancelRedisplay();
  if( sConfigure.zVariable && *sConfigure.zVariable ){
    Tcl_UntraceVar(interp, sConfigure.zVariable,
        TCL_TRACE_WRITES|TCL_TRACE_UNSETS|TCL_GLOBAL_ONLY,
        (Tcl_VarTraceProc*)ss::SsVarProc,
        (ClientData) this);
  }

  if( font ){
    Tk_FreeFont(font);
    font = 0;
  }
  delete pSelectionRect;

  delete [] sConfigure.zBg;
  delete [] sConfigure.zFg;
  delete [] sConfigure.zActiveBg;
  delete [] sConfigure.zActiveFg;
  delete [] sConfigure.zSelectBg;
  delete [] sConfigure.zSelectFg;
  delete [] sConfigure.zTitleBg;
  delete [] sConfigure.zTitleFg;

  delete [] sConfigure.zCursor;
  delete [] sConfigure.zFont;
  delete [] sConfigure.zResizeColor;
  delete [] sConfigure.zHResizeCursor;
  delete [] sConfigure.zVResizeCursor;
  delete [] sConfigure.zTitleCursor;
  delete [] sConfigure.zState;
  delete [] sConfigure.zIfDisabled;
  delete [] sConfigure.zVariable;
  delete [] sConfigure.zXScrollCmd;
  delete [] sConfigure.zYScrollCmd;

  delete [] zSs;
}

/*
** Activate the cell with the index pointed to by zIndex.
*/
PUBLIC void ss::Activate(char *zIndex){
  SsCell cellToActivate;  // The cell that will become active
  SsString iCursor;       // The insertion cursor for that cell, if any

  if( !zIndex ) return;

  if( *zIndex ){
    if( evalf("%s index %s", zSs, zIndex)!=TCL_OK ){
      return;
    }
    SsCell cellFromIndex(evalStrResult());

    if( !cellFromIndex.isnull() ){
      float x, y;
      if( sscanf(zIndex, "@%f,%f", &x, &y)==2 ){

        /* Get the highest or leftmost non-title cell closest to the index,
        ** if the index is in pixel coordinates.
        ** This mimics Excel behavior.
        */
        if( cellFromIndex.GetRow()<=0 ){
          cellFromIndex.SetRow(topleft.GetRow());
        }
        if( cellFromIndex.GetCol()<=0 ){
          cellFromIndex.SetCol(topleft.GetCol());
        }

        /* Get the insertion cursor point.
        */
        if( cellFromIndex.GetRow()>0 && cellFromIndex.GetCol()>0 ){
          iCursor += "@";
          fastcan.Eval("canvasx %g", x);
          iCursor += evalStrResult();
          iCursor += ",";
          fastcan.Eval("canvasy %g", y);
          iCursor += evalStrResult();
        }
      }

      if( cellFromIndex.GetRow()>0 && cellFromIndex.GetCol()>0 ){
        cellToActivate = cellFromIndex;
      }
    }
  }

  CommitActiveValue();

  /* Note that we do *not* check to see if the active cell is actually changing.
  ** We want to force any traces to fire.
  **
  ** Make it so there is no active cell, if there was one previously
  */
  SsCell prevActive = activeCell;
  if( !activeCell.isnull() ){
    activeCell.clear();
    fastcan.Focus(0);
    fastcan.DtagAll("active");
    fastcan.DtagAll("activecell");
    RecolorCell(prevActive);
  }

  /* Activate the cell to activate, if there is one.
  ** It may be necessary to create a text glyph on the fly if it doesn't
  ** already exist for the cell.
  */
  if( !cellToActivate.isnull() ){
    char *zCell = cellToActivate.render();
    if( strcmp(sConfigure.zState, "disabled")!=0 ){
      if( fastcan.TagToFirstId(zCell)==0 ){
        char zCellTag[64];
        sprintf(zCellTag, "C%s", zCell);
        int cellId = fastcan.TagToFirstId(zCellTag);
        if( cellId!=0 ) CreateTextGlyph(cellId, "");
      }
      fastcan.Focus(zCell);
      if( !iCursor.isnull() ){
        fastcan.Icursor(zCell, iCursor.render());
      } else if( !cellToActivate.equal(prevActive) ){
        fastcan.Icursor(zCell, "end");
      }
    }

    activeCell.set(cellToActivate.GetRow(), cellToActivate.GetCol());
    fastcan.AddTag("active", zCell);
    SsString cellTag("C"); cellTag += zCell;
    fastcan.AddTag("activecell", cellTag.render());
    RecolorCell(cellToActivate);
    SetWidgetVariable("active", GetWidgetVariable(cellToActivate.render()));
  } else {
    SetWidgetVariable("active", "");
  }
}

/*
** Return the bounding box that contains all of the canvas items
** referred to by the indices zIndex1 and zIndex2.
** The return is malloc'ed; the caller must be careful to free it.
*/
PUBLIC SsBbox *ss::Bbox(const char *zIndex1, const char *zIndex2){
  SsCell *pCell0 = Index(zIndex1);
  if( !pCell0 ){
    return 0;
  }
  SsCell *pCell1 = Index(zIndex2);
  if( !pCell1 ){
    delete pCell0;
    return 0;
  }

  /* Make sure to use the tag associated with the cell, not the cell text.
  */
  SsString cellTag0("C"); cellTag0.append(pCell0->render());
  SsString cellTag1("C"); cellTag1.append(pCell1->render());
  delete pCell0;
  delete pCell1;

  SsSetOfString *pTagSet = new SsSetOfString;
  pTagSet->insert(cellTag0.render());
  pTagSet->insert(cellTag1.render());
  SsBbox *pBbox = fastcan.Bbox(pTagSet);

  pTagSet->reset();
  delete pTagSet;

  if( !pBbox || pBbox->isnull() ){
    delete pBbox;
    return 0;
  }
  return pBbox;
}

/*
** Return the bounding box for the given cell.
** The return is malloc'ed.
*/
PRIVATE SsBbox *ss::Bbox(SsCell &cell){

  /* Make sure to use the tag associated with the cell, not the cell text.
  */
  SsString cellTag("C"); cellTag.append(cell.render());
  SsSetOfString *pTagSet = new SsSetOfString;
  pTagSet->insert(cellTag.render());
  SsBbox *pBbox = fastcan.Bbox(pTagSet);

  pTagSet->reset();
  delete pTagSet;

  if( !pBbox || pBbox->isnull() ){
    delete pBbox;
    return 0;
  }
  return pBbox;
}

/*
** Configure all cells to have their default foreground color,
** default background color, or both.  If zOption is NULL, then both defaults
** are set to defaults.  If zOption is "-fg", then just the foreground is
** set to default; if zOption is "-bg" then just the background is set to
** default.
*/
PUBLIC void ss::CellConfigureDefault(char *zOption){
  if( zOption ){
    if( strcmp(zOption, "-fg")==0 ){
      aFg.reset();
      RecolorAllForegrounds();
    } else if( strcmp(zOption, "-bg")==0 ){
      aBg.reset();
      RecolorAllBackgrounds();
    }
  } else {
    aFg.reset();
    aBg.reset();
    RecolorAllCells();
  }
}

/*
** Configure the cell's "-fg" or "-bg" option.
*/
PUBLIC void ss::CellConfigure(char *zCell, int objc, Tcl_Obj *const* objv){
  SsCell cell(zCell);
  for( int i = 0; i < objc; i+=2 ){
    char *zOption = Tcl_GetStringFromObj(objv[i], 0);
    if( strcmp(zOption, "-fg")==0 ){
      char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);
      if( zValue && !*zValue ){
        zValue = 0;
      }
      char *zOldValue = (char *)aFg.find(zCell);
      if( !zOldValue && !zValue ) continue;
      if( zOldValue && zValue && strcmp(zValue, zOldValue)==0 ) continue;
      if( zOldValue ) {
        delete [] zOldValue;
      }
      aFg.insert(zCell, zValue);
      RecolorCellFg(cell);
    } else if( strcmp(zOption, "-bg")==0 ){
      char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);
      if( zValue && !*zValue ){
        zValue = 0;
      }
      char *zOldValue = (char *)aBg.find(zCell);
      if( !zOldValue && !zValue ) continue;
      if( zOldValue && zValue && strcmp(zValue, zOldValue)==0 ) continue;
      if( zOldValue ) {
        delete [] zOldValue;
      }
      aBg.insert(zCell, zValue);
      RecolorCellBg(cell);
    }
  }
}

/*
** Configure the foreground color of all cells in the "objv" list.
** The foreground color is pointed to by zFg.
*/
PUBLIC void ss::CellConfigureFg(char *zFg, int objc, Tcl_Obj *const* objv){
  int bDisplayPending = (displayProperties & DISPLAY_PENDING);
  if( zFg && !*zFg ){
    zFg = 0;
  }
  for( int i=0; i<objc; i++ ){
    char *zCell = Tcl_GetStringFromObj(objv[i], 0);
    int row, col;
    if( sscanf(zCell, "%d,%d", &row, &col)!=2 || row<0 || col<0 ) continue;
    char *zOldFg = (char *)aFg.find(zCell);
    if( !zOldFg && !zFg ) return;
    if( zOldFg && zFg && strcmp(zFg, zOldFg)==0 ) continue;
    if( zOldFg ) {
      delete [] zOldFg;
    }
    aFg.insert(zCell, zFg);
    if( !bDisplayPending ){
      SsCell cell(zCell);
      fastcan.Itemconfigure(zCell, "-fill", GetCellFg(cell));
    }
  }
}

/*
** Configure the background color of all cells in the "objv" list.
** The background color is pointed to by zBg.
*/
PUBLIC void ss::CellConfigureBg(char *zBg, int objc, Tcl_Obj *const* objv){
  int bDisplayPending = (displayProperties & DISPLAY_PENDING);
  if( zBg && !*zBg ){
    zBg = 0;
  }
  for( int i=0; i<objc; i++ ){
    char *zCell = Tcl_GetStringFromObj(objv[i], 0);
    int row, col;
    if( sscanf(zCell, "%d,%d", &row, &col)!=2 || row<0 || col<0 ) continue;
    char *zOldBg = (char *)aBg.find(zCell);
    if( !zOldBg && !zBg ) return;
    if( zOldBg && zBg && strcmp(zBg, zOldBg)==0 ) continue;
    if( zOldBg ) {
      delete [] zOldBg;
    }
    aBg.insert(zCell, zBg);
    if( !bDisplayPending ){
      char zCellTag[128];
      *zCellTag = 'C';
      strncpy(zCellTag+1, zCell, (sizeof zCellTag)-1);
      SsCell cell(zCell);
      fastcan.Itemconfigure(zCellTag, "-fill", GetCellBg(cell));
    }
  }
}

/*
** Return the value of the option referenced by zOption.
** If there is no such option, return 0.
*/
PUBLIC char *ss::Cget(const char *zOption){
  return sConfigure.GetOptionValue(zOption);
}

/*
** "Clone" the part of the spreadsheet within the range of rows
** given by "from" and "to", and write the spreadsheet on "zNewCanvas".
** "rowMag" is the magnification for any row.
*/
PUBLIC void ss::Clone(int from, int to, char *zNewCanvas, float rowMag){
  if( nDisplayedRows==0 ) return;

  if( (from<1 && to<1) || (from>=nDisplayedRows && to>=nDisplayedRows) ){
    from = 0;
    to = 0;
  } else {
    if( from < 1 ){
      from = 1;
    }
    if( from >= nDisplayedRows ){
      from = nDisplayedRows - 1;
    }
    if( to < 1 ){
      to = 1;
    }
    if( to >= nDisplayedRows ){
      to = nDisplayedRows - 1;
    }
    if( from > to ){
      int tmp = from;
      from = to;
      to = tmp;
    }
  }
  CopySpreadsheet(from, to, zNewCanvas, rowMag);
}

/*
** Configure the spreadsheet widget from the OPTION/VALUE pairs in
** the objv list.  Return a Tcl result.
*/
PUBLIC int ss::Configure(int objc, Tcl_Obj *const* objv){

  static const char *SS_config_strs[] = {
    "-activebackground", "-activebg",           "-activefg",            
    "-activeforeground", "-background",         "-bg",
    "-cols",             "-colwidth",           "-cursor",
    "-expandcol",        "-fg",                 "-font",
    "-foreground",       "-height",             "-hidecolor",
    "-hresizecursor",    "-ifdisabled",         "-maxrows",
    "-mincolwidth",      "-minrowheight",       "-resizecolor",
    "-rowheight",        "-rows",               "-selectbackground",
    "-selectbg",         "-selectfg",           "-selectforeground",
    "-state",            "-titlebackground",    "-titlebg",
    "-titlefg",          "-titleforeground",    "-titlerowheight",
    "-titlecolwidth",    "-variable",           "-vresizecursor",
    "-watermark",        "-width",              "-xscrollcommand",
    "-yscrollcommand",
    0,
  };
  enum SS_config_index {
    SS_CFG_ACTIVEBG,       SS_CFG_ACTIVEBG2,       SS_CFG_ACTIVEFG,
    SS_CFG_ACTIVEFG2,      SS_CFG_BG,              SS_CFG_BG2,
    SS_CFG_COLS,           SS_CFG_COLWIDTH,        SS_CFG_CURSOR,
    SS_CFG_EXPANDCOL,      SS_CFG_FG,              SS_CFG_FONT,
    SS_CFG_FG2,            SS_CFG_HEIGHT,          SS_CFG_HIDECOLOR,
    SS_CFG_HRESIZECURSOR,  SS_CFG_IFDISABLED,      SS_CFG_MAXROWS,
    SS_CFG_MINCOLWIDTH,    SS_CFG_MINROWHEIGHT,    SS_CFG_RESIZECOLOR,
    SS_CFG_ROWHEIGHT,      SS_CFG_ROWS,            SS_CFG_SELECTBG,
    SS_CFG_SELECTBG2,      SS_CFG_SELECTFG,        SS_CFG_SELECTFG2,
    SS_CFG_STATE,          SS_CFG_TITLEBG,         SS_CFG_TITLEBG2,
    SS_CFG_TITLEFG,        SS_CFG_TITLEFG2,        SS_CFG_TITLEROWHEIGHT,
    SS_CFG_TITLECOLWIDTH,  SS_CFG_VARIABLE,        SS_CFG_VRESIZECURSOR,
    SS_CFG_WATERMARK,      SS_CFG_WIDTH,           SS_CFG_XSCROLLCOMMAND,
    SS_CFG_YSCROLLCOMMAND,
  };

  /* Do a prepass that verifies that the values have the right type
  */
  Tk_Font new_font = 0;
  int index;
  for( int i=0; i<objc; i+=2 ){
    if( Tcl_GetIndexFromObj(interp, objv[i], SS_config_strs, "option", 0,
        &index) ){
      continue;
    }
    switch( (enum SS_config_index) index ){
      case SS_CFG_COLS:
      case SS_CFG_COLWIDTH:
      case SS_CFG_EXPANDCOL:
      case SS_CFG_HEIGHT:
      case SS_CFG_MAXROWS:
      case SS_CFG_MINCOLWIDTH:
      case SS_CFG_ROWHEIGHT:
      case SS_CFG_ROWS:
      case SS_CFG_TITLECOLWIDTH:
      case SS_CFG_TITLEROWHEIGHT:
      case SS_CFG_WIDTH:
      {
        char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);
        int j;
        if( sscanf(zValue, "%d", &j)!=1 ){
          char *zOption = Tcl_GetStringFromObj(objv[i], 0);
          Tcl_AppendResult(interp, "Value: ", zValue, " for option ", zOption,
              " not an integer", 0);
          return TCL_ERROR;
        }
        break;
      }
      case SS_CFG_FONT:
      {
        char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);
        Tk_Window tkwin = Tk_NameToWindow(interp, zSs, Tk_MainWindow(interp));
        new_font = Tk_GetFont(interp, tkwin, zValue);
        if( !new_font ){
          return TCL_ERROR;
        }
        break;
      }
      case SS_CFG_STATE:
      {
        char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);
        if( strcmp(zValue, "normal")!=0 && strcmp(zValue, "disabled")!=0 ){
          char *zOption = Tcl_GetStringFromObj(objv[i], 0);
          Tcl_AppendResult(interp, "Value: ", zValue, " for option ", zOption,
              " not \"normal\" or \"disabled\"", 0);
          return TCL_ERROR;
        }
        break;
      }
      case SS_CFG_IFDISABLED:
      {
        char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);
        if( strcmp(zValue, "all")!=0 && strcmp(zValue, "table")!=0 ){
          char *zOption = Tcl_GetStringFromObj(objv[i], 0);
          Tcl_AppendResult(interp, "Value: ", zValue, " for option ", zOption,
              " not \"all\" or \"table\"", 0);
          return TCL_ERROR;
        }
        break;
      }

      default:
        break;
    }
  }

  int bRedisplay = 0;   // nonzero if the entire spreadsheet must be redisplayed
  for( int i=0; i<objc; i+=2 ){
    if( Tcl_GetIndexFromObj(interp, objv[i], SS_config_strs, "option", 0,
        &index) ){
      continue;
    }
    char *zOption = Tcl_GetStringFromObj(objv[i], 0);
    char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);

    /* Get the old value of the option.
    ** We must make a copy of the old option because we'll be changing the
    ** value below.
    */
    SsString oldValue(sConfigure.GetOptionValue(zOption));
    char *zOldValue = oldValue.render();
    Trace(TRACE_COMMAND, "CONFIG OPTION: %s VALUE: %s\n", zOption, zValue);

    if( strcmp(zValue, zOldValue)==0 ){  // no change
      continue;
    }

    /* Save the options.
    */
    sConfigure.SetOptionValue(zOption, zValue);

    switch( (enum SS_config_index) index ){
      case SS_CFG_CURSOR:
      case SS_CFG_XSCROLLCOMMAND:
      case SS_CFG_YSCROLLCOMMAND:  // these are options for the canvas
        fastcan.Eval("config %s \"%q\"", zOption, zValue);
        break;

      case SS_CFG_MAXROWS:
        InitTitleColumn();
        bRedisplay++;
        break;

      case SS_CFG_VARIABLE:
        if( *zOldValue ){  // Remove the old variable trace
          Tcl_UntraceVar(interp, zOldValue,
              TCL_TRACE_WRITES|TCL_TRACE_UNSETS|TCL_GLOBAL_ONLY,
              (Tcl_VarTraceProc*)ss::SsVarProc,
              (ClientData) this);
        }
        if( zValue && *zValue ){ // Set the new variable trace
          Tcl_TraceVar(interp, zValue, 
              TCL_TRACE_WRITES|TCL_TRACE_UNSETS|TCL_GLOBAL_ONLY,
              (Tcl_VarTraceProc*)ss::SsVarProc,
              (ClientData) this);
        }
        bDirtyCell++;

        /* fall through */

      default:
        bRedisplay++;
        break;
    }
  }
  if( new_font ){
    if( font ){
      Tk_FreeFont(font);
    }
    font = new_font;
  }

  if( bRedisplay ){
    Display(0, 0);
  }
  return TCL_OK;
}

/*
** Initialize the title column containing the row data.
** Done every time the maximum number of rows is changed.
*/
PRIVATE void ss::InitTitleColumn(){
  if( sConfigure.zVariable && *sConfigure.zVariable ){
    for( int row=1; row < sConfigure.maxRows; row++ ){
      char zCell[35];
      sprintf(zCell, "%d,0", row);
      if( !GetWidgetVariable(zCell) ){  // check to avoid trace firing
        char zRow[33];
        sprintf(zRow, "%d", row);
        SetWidgetVariable(zCell, zRow);
      }
    }
  }
}

/*
** Return a pointer to an SsTclList containing a sorted list of cells in the 
** selection.
** The return is malloc'ed; the caller must free it.
*/
PUBLIC SsTclList *ss::GetCurrentSelection(){

  /* If the widget variable is defined, then its "selected" component
  ** contains the current selection.  If not, the selection must
  ** be calculated from canvas tags.  Getting it from the variable is
  ** much faster than getting it from the canvas tags.
  */
  SsTclList *pSelection = new SsTclList(interp, GetWidgetVariable("selected"));
  if( pSelection->isnull() ){
    delete pSelection;
    pSelection = fastcan.GetSortedSelection();
  } else {

    /* Handle the case where the "selected" element's value is "all".
    */
    char *zValue = pSelection->render();
    if( *zValue=='a' && strcmp(zValue, "all")==0 ){
      pSelection->set("");
      for( int row = 1; row<nDisplayedRows; row++ ){
        char zCell[66];              // Safe for 2 32-bit integers, plus comma
        sprintf(zCell, "%d,", row);
        char *zCol = zCell+strlen(zCell);
        for( int col = 1; col<sConfigure.cols; col++ ){
          if( col>=10 ){  // max column==99
            zCol[0] = '0' + col/10;
            zCol[1] = '0' + col%10;
            zCol[2] = 0;
          } else {
            zCol[0] = '0' + col;
            zCol[1] = 0;
          }
          pSelection->append(zCell);
        }
      }
    }
  }
  return pSelection;
}

/*
** Return a pointer to an SsTclList containing the cell with the
** current selection, the index in the cell that defines the start of
** the selection, and the index defining the end of the selection.
** Return 0 if there is no current text selection.
** The return value is malloc'ed; the caller must delete it.
*/
PUBLIC SsTclList *ss::GetCurrentTextSelection(){
  int id, from, to;
  if( !FindTextSelection(&id, &from, &to) ){
    return 0;
  }
  SsCell *pCell = fastcan.IdToCell(id);
  if( !pCell ){
    return 0;
  }
  SsTclList *pList = new SsTclList(interp, pCell->render());
  pList->append(from);
  pList->append(to);
  return pList;
}

/*
** Return the current value of the active cell.
** Return 0 if there is no active cell.
** The return is malloc'ed; the caller must free it.
*/
PUBLIC SsString *ss::GetCurrentValue(){
  return fastcan.Itemcget("active", "-text");
}

/*
** Set the current value of the active cell to zValue.
** If there is no currently active cell, nothing happens.
*/
PUBLIC void ss::SetCurrentValue(const char *zValue){
  int id = fastcan.FindFirst("activecell");
  if( id==0 ) return;
  SsCell *pCell = fastcan.IdToCell(id);
  if( !pCell ) return;
  fastcan.SetText(pCell->render(), zValue ? zValue : "", this);
}

/*
** Delete characters from the given cell,
** from index zFrom to index zTo.
*/
PUBLIC void ss::Delete(SsCell &cell, char *zFrom, char *zTo){
  if( cell.isnull() ) return;
  fastcan.Dchars(cell.render(), zFrom, zTo);
  if( cell.equal(activeCell) ){
    SetActiveVar();
  }
  RecolorCell(cell);
}

/*
** Set the insertion cursor on the currently active cell to be at "pos".
** Does nothing if the spreadsheet is disabled.
*/
PUBLIC void ss::SetIcursor(int pos){
  if( pos>=0 && strcmp(sConfigure.zState, "normal")==0 ){
    fastcan.Icursor("active", pos);
  }
}

/*
** This proc returns a pointer to an SsString representation of the cell 
** with the given zIndex.
** zRowCol may be either "row", "col", or 0.
** If "row" or "col" is specified, then only the row or column index
** is returned.
** Returns 0 if there is no cell with the given index.
** The return is malloc'ed; the caller must free it.
*/
PUBLIC SsString *ss::GetCellComponent(const char *zIndex, const char *zRowCol){
  SsCell *pCell = Index(zIndex);
  if( !pCell ){
    return 0;
  }
  SsString *pResult = new SsString();

  if( !zRowCol ){
    pResult->append(pCell->render());
  } else if( *zRowCol=='r' && strcmp(zRowCol, "row")==0 ){
    (*pResult) += pCell->GetRow();
  } else if( *zRowCol=='c' && strcmp(zRowCol, "col")==0 ){
    (*pResult) += pCell->GetCol();
  }
  delete pCell;
  return pResult;
}

/*
** Return the cell referenced by the table index zIndex.
** Return 0 if there is none.
** The return is malloc'ed; the caller must free it.
*/
PUBLIC SsCell *ss::Index(const char *zIndex){
  float x,y;
  int row, col;

  if( sscanf(zIndex, "@%f,%f", &x, &y)==2 ){
    fastcan.Eval("canvasx %g", x);
    SsString cx(evalStrResult());  // can be a floating point number
    fastcan.Eval("canvasy %g", y);
    SsString cy(evalStrResult());  // can be a floating point number
    ShrinkBgrect();

    int prev = -1;
    int first = -1;
    SsCell *pCell = 0;
    while( 1 ){
      fastcan.Eval("find closest %s %s 0 %d", cx.render(), cy.render(), prev);
      int id = evalIntResult();
      if( id==0 ) break;
      if( first<0 ){
        first = id;
      } else if( first==id ){  // finished the cycle
        break;
      }
      SsCell *p = fastcan.IdToCell(id);
      if( p ){
        pCell = new SsCell(p);
        break;
      }
      prev = id;
    }
    RestoreBgrect();
    return pCell;

  } else if( sscanf(zIndex, "%d,%d", &row, &col)==2 ){
    return new SsCell(row, col);

  } else if( *zIndex=='a' && strcmp(zIndex, "active")==0 ){
    return new SsCell(activeCell);

  } else if( *zIndex=='a' && strcmp(zIndex, "anchor")==0 ){
    return new SsCell(selectionAnchor);

  } else if( *zIndex=='e' && strcmp(zIndex, "end")==0 ){
    return new SsCell(nDisplayedRows - 1, sConfigure.cols - 1);

  } else if( *zIndex=='o' && strcmp(zIndex, "origin")==0 ){
    return new SsCell(1, 1);

  } else if( *zIndex=='t' && strcmp(zIndex, "topleft")==0 ){
    return new SsCell(topleft);

  } else if( *zIndex=='b' && strcmp(zIndex, "bottomleft")==0 ){
    int brow = GetRowAtDistance(topleft.GetRow(), GetScrollableHeight());
    if( brow>=nDisplayedRows ){
      brow = nDisplayedRows - 1;
    }
    return new SsCell(brow, topleft.GetCol());
  }
  return 0;
}

/*
** Configure a set of heights from the even-numbered list of
** (row, height) pairs in objv.
** This will cause the "heights" widget variable to be reset.
** It will also cause the spreadsheet to be redisplayed if any row height
** actually changed.
*/
PUBLIC void ss::ConfigureHeights(int objc, Tcl_Obj *const* objv){
  int bChanged = 0;
  for( int i=0; i < objc; i+=2 ){
    int row = atoi(Tcl_GetStringFromObj(objv[i], 0));
    char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);
    int height = 0;
    if( *zValue!='d' || strcmp(zValue, "default")!=0 ){
      height = atoi(zValue);
    }
    if( ConfigureHeight(row, height) ){
      bChanged++;
    }
  }
  SsTclList *pHeights = GetAllHeights();
  fromSs = 1;
  SetWidgetVariable("heights", pHeights->render());
  fromSs = 0;
  delete pHeights;

  if( bChanged ){
    Display(0, 0, 0);
  }
}

/*
** This method configures the height of "row" to be "height."
** But if "height" is 0, then the row is given the default height.
**
** Return nonzero if the width for "row" was changed.
*/
PRIVATE int ss::ConfigureHeight(int row, int height){
  if( row<1 ){
    return 0;
  }
  if( height!=0 && height<sConfigure.minRowHeight ){
    height = sConfigure.minRowHeight;
  }
  if( height==sConfigure.rowHeight ){
    height = 0;
  }
  int curHeight = (int) aHeight[(void *)row];
  if( curHeight==height ){
    return 0;
  }
  aHeight.insert((void *)row, (void *)height);  // handles default automatically
  return 1;
}

/*
** Reset all configured heights to have their defaults.
** Set the widget variable's "heights" element.
** Schedule a redisplay of the widget.
*/
PUBLIC void ss::ResetConfiguredHeights(){
  if( aHeight.count()==0 ){
    return;
  }
  aHeight.reset();
  fromSs = 1;
  SetWidgetVariable("heights", "");
  fromSs = 0;
  Display();
}

/*
** Return the configured height of the given row.
*/
PUBLIC int ss::GetConfiguredHeight(int row){
  int ht = (int) aHeight[(void *)row];
  if( ht==0 ){
    if( row==0 ){
      ht = (currentTitleRowHeight==-1) ? 
          sConfigure.titleRowHeight : currentTitleRowHeight;
    } else {
      ht = sConfigure.rowHeight;
    }
  }
  return ht;
}

/*
** Return a pointer to a SsString object that contains a sorted list of
** all configured heights (except possibly for the title row).
** The caller must free the SsString object.
*/
PUBLIC SsTclList *ss::GetAllHeights(){

  /* First, get the list of rows and sort them.
  */
  SsTclList *pRows = new SsTclList(interp);
  for( SsArrayElem *pElem = aHeight.first(); pElem; pElem = pElem->next() ){
    pRows->append((int)(pElem->key()));
  }
  evalf("lsort -integer \"%s\"", pRows->render());
  pRows->set((char *)evalStrResult());

  /* Now create the even-numbered list of rows and heights, sorted by row.
  */
  SsTclList *pHeights = new SsTclList(interp);
  for( char *zRow = pRows->first(); zRow; zRow = pRows->next() ){
    int row = atoi(zRow);
    if( row<1 ) continue;
    int height = (int)aHeight[(void *)row];
    pHeights->append(row);
    pHeights->append(height);
  }

  delete pRows;
  return pHeights;
}

/*
** Configure a set of widths from the even-numbered list of
** (column, width) pairs in objv.
** This will cause the "widths" widget variable to be reset.
** It will also cause the spreadsheet to be redisplayed if any column width
** actually changed.
*/
PUBLIC void ss::ConfigureWidths(int objc, Tcl_Obj *const* objv){
  int bChanged = 0;
  for( int i=0; i < objc; i+=2 ){
    int col = atoi(Tcl_GetStringFromObj(objv[i], 0));
    char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);
    int width = 0;
    if( strcmp(zValue, "default")!=0 ){
      width = atoi(zValue);
    }
    if( ConfigureWidth(col, width) ){
      bChanged++;
    }
  }

  SsTclList *pWidths = GetAllWidths();
  fromSs = 1;
  SetWidgetVariable("widths", pWidths->render());
  fromSs = 0;
  delete pWidths;

  if( bChanged ){
    Display(0, 0, 0);
  }
}

/*
** This method configures the width of "col" to be "width".
** But if "width" is 0, then the row is given the default width.
**
** Return nonzero if the width for "col" was changed.
*/
PRIVATE int ss::ConfigureWidth(int col, int width){
  int minWidth = GetMinimumWidth(col);
  if( width!=0 && width<minWidth ){
    width = minWidth;
  }
  if( width==sConfigure.colWidth ){
    width = 0;
  }
  int curWidth = (int) aWidth[(void *)col];
  if( curWidth==width ){
    return 0;
  }
  aWidth.insert((void *)col, (void *)width);  // handles default automatically
  ResetCurrentWidth();
  return 1;
}

/*
** This method toggles the hidden state of column "col".
** Nothing happens if "col" does not exist.
*/
PRIVATE void ss::ToggleHide(int col){
  if( col<0 || col>=sConfigure.cols ) return;
  int hidden = (int) aHidden[(void *)col];
  hidden ^= 1;
  aHidden.insert((void *)col, (void *)hidden);
  ResetCurrentWidth();
  Display(0, 0);
}

/*
** Reset all configured widths to have their defaults.
** Set the widget variable's "widths" element.
** Schedule a redisplay of the widget.
*/
PUBLIC void ss::ResetConfiguredWidths(){
  if( aWidth.count()==0 ){
    return;
  }
  aWidth.reset();
  ResetCurrentWidth();

  fromSs = 1;
  SetWidgetVariable("widths", "");
  fromSs = 0;
  Display();
}

/*
** This method will reset the "currentTotalWidth" member element.
** It should be called whenever a column width changes.
*/
PRIVATE void ss::ResetCurrentWidth(){
  currentTotalWidth = 0;
  for( int col=0; col<sConfigure.cols; col++ ){
    currentTotalWidth += GetColumnWidth(col);
  }
}

/*
** Return the configured width of the given column.
*/
PUBLIC int ss::GetConfiguredWidth(int col){
  int wid = (int) aWidth[(void *)col];
  if( wid==0 ){
    if( col==0 ){
      wid = sConfigure.titleColWidth;
    } else {
      wid = sConfigure.colWidth;
    }
  }
  return wid;
}

/*
** This proc returns 1 if "col" is hidden and 0 otherwise.
*/
PUBLIC int ss::GetHidden(int col){
  return (int) aHidden[(void *)col];
}

/*
** This is like GetConfiguredWidth(), but instead of returning the 
** configured width, it will return 0 if the given column is hidden.
*/
PUBLIC int ss::GetColumnWidth(int col){
  int hidden = (int) aHidden[(void *)col];
  return hidden? 0 : GetConfiguredWidth(col);
}

/*
** Return a pointer to a SsString object that contains a sorted list of
** all configured widths.  The caller must free the SsString object.
*/
PUBLIC SsTclList *ss::GetAllWidths(){

  /* First, get the list of columns and sort them.
  */
  SsTclList *pCols = new SsTclList(interp);
  for( SsArrayElem *pElem = aWidth.first(); pElem; pElem = pElem->next() ){
    pCols->append((int)(pElem->key()));
  }
  evalf("lsort -integer \"%s\"", pCols->render());
  pCols->set((char *)evalStrResult());

  /* Now create the even-numbered list of columns and widths, sorted by column
  */
  SsTclList *pWidths = new SsTclList(interp);
  for( char *zCol = pCols->first(); zCol; zCol = pCols->next() ){
    int col = atoi(zCol);
    int width = (int)aWidth[(void *)col];
    pWidths->append(col);
    pWidths->append(width);
  }

  delete pCols;
  return pWidths;
}

/*
** Return the minimum width of the given column, which is the maximum
** of the "-mincolwidth" value and the per-column minimum width.
**
** If there is no per-column minimum, this returns the "-mincolwidth"
** configured value.
*/
PUBLIC int ss::GetMinimumWidth(int col){
  int minWidth = (int)aMinWidth[(void *)col];
  if( minWidth<sConfigure.minColWidth ){
    minWidth = sConfigure.minColWidth;
  }
  return minWidth;
}

/*
** Set the minimum width of the given column.
** This may be set below the configured minimum in "-mincolwidth".
** But when the minimum is actually calculated (see GetMinimumWidth() above),
** the greater of the two are used.
*/
PUBLIC void ss::SetMinimumWidth(int col, int width){
  aMinWidth.insert((void *)col, (void *)width);
}

/*
** Return a pointer to an SsTclList containing the width and height of the
** canvas viewport.
** The return is malloc'ed; the caller must free it.
*/
PUBLIC SsTclList *ss::GetViewport(){
  SsTclList *pViewport = new SsTclList(interp);
  fastcan.Eval("cget -width");
  pViewport->append(evalIntResult());
  fastcan.Eval("cget -height");
  pViewport->append(evalIntResult());
  return pViewport;
}

/*
** Insert the string zString at the current insertion cursor for the active
** cell.  If there is an intra-cell selection, it is removed first.
** If there is no active cell, nothing happens.
**
** As a result of this proc, the height of the row containing the
** active cell is recomputed, and the active cell is reconfigured.
*/
PUBLIC void ss::InsertActive(const char *zString){
  if( strcmp(sConfigure.zState, "disabled")==0 ) return;
  if( !zString || !*zString || activeCell.isnull() ) return;

  int activeId = fastcan.FindFirst("active");
  int id, from, to;
  if( FindTextSelection(&id, &from, &to) && from>=0 && to>=0 && id==activeId ){
    fastcan.Dchars("active", from, to);
    fastcan.Eval("select clear");
  }

  fastcan.InsertActive(zString, this);
  ResetHeight(activeCell);
  SetActiveVar();
  RecolorCell(activeCell);
  See(activeCell.render());
}

/*
** This method will scroll the canvas so that the first cell that
** is at the given zIndex will be visible.
*/
PUBLIC void ss::See(const char *zIndex){
  SsCell *pCell = Index(zIndex);
  if( !pCell ){
    return;
  }
  MakeCellVisible(*pCell);
  delete pCell;
}

/*
** Set the selection anchor to be the cell whose name is zCell
*/
PUBLIC void ss::SetSelectionAnchor(const char *zCell){
  selectionAnchor.set(zCell);
}

/*
** Clear the selection in the rectangle of cells defined
** from zFirstIndex to zLastIndex.
*/
PUBLIC void ss::SelectionClearCells(const char *zFirstIndex, 
    const char *zLastIndex){
  if( strcmp(zFirstIndex, "all")==0 ){
    DeselectAllCells();
  } else {
    SsCell *pFirst = Index(zFirstIndex);
    if( !pFirst ) return;
    SsCell *pLast = Index(zLastIndex);
    if( !pLast ){
      delete pFirst;
      return;
    }
    int lowRow = min(pFirst->GetRow(), pLast->GetRow());
    int hiRow = max(pFirst->GetRow(), pLast->GetRow());
    int lowCol = min(pFirst->GetCol(), pLast->GetCol());
    int hiCol = max(pFirst->GetCol(), pLast->GetCol());
    delete pFirst;
    delete pLast;

    for( int row = lowRow; row <= hiRow; row++ ){
      for( int col = lowCol; col <= hiCol; col++ ){
        SsCell cell(row, col);
        fastcan.DeselectCell(cell.render());
        DeselectCell(cell);
      }
    }
  }
  SsTclList *pNew = fastcan.GetSortedSelection();
  SetSelectedVariable(pNew);
  delete pNew;
  DrawSelectionRectangle();
}

/*
** Delete all of the cell data in the current selection.
*/
PUBLIC void ss::SelectionClearData(){
  SsTclList *pSelected = GetCurrentSelection();
  if( sConfigure.zVariable && *sConfigure.zVariable ){
    for( char *zCell = pSelected->first(); zCell; zCell = pSelected->next() ){
      if( Tcl_GetVar2(interp, sConfigure.zVariable, zCell, TCL_GLOBAL_ONLY) ){
        Tcl_UnsetVar2(interp, sConfigure.zVariable, zCell, TCL_GLOBAL_ONLY);
      }
    }
  } else {
    for( char *zCell = pSelected->first(); zCell; zCell = pSelected->next() ){
      fastcan.Dchars(zCell, "0", "end");
    }
  }
  delete pSelected;
}

/*
** Return nonzero if the set of selected cells includes the cell
** referenced by zIndex.
*/
PUBLIC int ss::SelectionIncludes(const char *zIndex){
  SsCell *pCell = Index(zIndex);
  if( !pCell ){
    return 0;
  }
  int bIncludes = fastcan.CellIsSelected(pCell->render());
  delete pCell;
  return bIncludes;
}

/*
** Return nonzero if there is a selection present.
*/
PUBLIC int ss::SelectionPresent(){
  SsTclList *pList = GetCurrentSelection();
  int bPresent = !pList->isnull();
  delete pList;
  return bPresent;
}

/*
** Set the selection to all cells in the rectangle swept out by
** the first cell referenced by zFirstIndex, to the last cell
** referenced by zLastIndex.  If bOnly is nonzero, then the selection is
** set to just those cells; otherwise, the new cells are added to the
** existing selection.
*/
PUBLIC void ss::SelectionSet(int bOnly, const char *zFirstIndex, 
    const char *zLastIndex){
  if( strcmp(zFirstIndex, "all")==0 ){
    fastcan.SelectAll(nDisplayedRows, sConfigure.cols);
    ColorAllSelected();
  } else {
    if( bOnly ){
      DeselectAllCells();
    }
    SsCell *pFirst = Index(zFirstIndex);
    if( !pFirst ) return;
    SsCell *pLast = Index(zLastIndex);
    if( !pLast ){
      delete pFirst;
      return;
    }
    int lowRow = min(pFirst->GetRow(), pLast->GetRow());
    int hiRow = max(pFirst->GetRow(), pLast->GetRow());
    int lowCol = min(pFirst->GetCol(), pLast->GetCol());
    int hiCol = max(pFirst->GetCol(), pLast->GetCol());
    if( hiRow<=0 || hiCol<=0 || 
        lowRow>=nDisplayedRows || lowCol>=sConfigure.cols ){
      return;
    }
    lowRow = max(lowRow, 1);
    lowCol = max(lowCol, 1);
    hiRow = min(hiRow, nDisplayedRows-1);
    hiCol = min(hiCol, sConfigure.cols-1);

    for( int row = lowRow; row <= hiRow; row++) {
      for( int col = lowCol; col <= hiCol; col++ ){
        SsCell cell(row, col);
        SelectionSetOneCell(cell);
      }
    }
  }
  SsTclList *pNew = fastcan.GetSortedSelection();
  SetSelectedVariable(pNew);
  delete pNew;
  DrawSelectionRectangle();
}

/*
** This method sets the selection to the cells referenced by the list
** of Tcl objects in ppObjv.  objc is the total number of cells to select.
*/
PUBLIC void ss::SelectionSetlist(int objc, Tcl_Obj **ppObjv){
  for( int i=0; i<objc; i++ ){
    SsCell cell(Tcl_GetString(ppObjv[i]));
    if( cell.GetRow()>0 && cell.GetRow()<nDisplayedRows &&
      cell.GetCol()>0 && cell.GetCol()<sConfigure.cols ){
      SelectionSetOneCell(cell);
    }
  }
  SsTclList *pNew = fastcan.GetSortedSelection();
  SetSelectedVariable(pNew);
  delete pNew;
  DrawSelectionRectangle();
}

/*
** Set the value of the "selected" element in the widget variable so
** that the elements of pSelection, each representing a cell, are selected.
** If pSelection represents every cell in the spreadsheet, 
** then the "selected" element will be set to "all".
** Doing things this way will save memory when the spreadsheet is very large.
**
** PRECONDITION:  The cells in the pSelection list are sorted in row-major
** order.
*/
PRIVATE void ss::SetSelectedVariable(SsTclList *pSelection) {
  if( !sConfigure.zVariable || !*sConfigure.zVariable ) return;
  int col = 1;
  int row = 1;
  int isAll = 1;
  for( char *zCell = pSelection->first(); zCell; zCell = pSelection->next() ){
    int nrow, ncol;
    if( sscanf(zCell, "%d,%d", &nrow, &ncol)!=2 || nrow!=row || ncol!=col ){
      isAll = 0;
      break;
    }
    col++;
    if( col==sConfigure.cols ){
      col = 1;
      row++;
    }
  }
  if( isAll && row==nDisplayedRows && col==1 ){
    SetWidgetVariable("selected", "all");
  } else {
    SetWidgetVariable("selected", pSelection->render());
  }
}

/*
** Set the selection for the single "cell".
*/
PRIVATE void ss::SelectionSetOneCell(SsCell &cell){
  fastcan.SelectCell(cell.render());
  SelectCell(cell);
}

/*
** This method sets the cell whose name is zCell so that
** it displays the given value pointed to by zValue.
*/
PUBLIC void ss::Set(char *zCell, char *zValue){
  if( sConfigure.zVariable && *sConfigure.zVariable ){
    SetWidgetVariable(zCell, zValue);
  } else {
    fastcan.SetText(zCell, zValue, this);
  }
}

/*
** This method will set the intra-cell (text) selection.
*/
PUBLIC void ss::SetTextSelection(const char *zIndex, int from, int to){
  SsCell *pCell = Index(zIndex);
  if( !pCell ) return;
  if( from<0 || to<0 ) return;

  if( to < from ){
    int tmp=to;
    to=from;
    from=tmp;
  }
  fastcan.Eval("select clear");
  fastcan.SelectFrom(pCell->render(), from);
  fastcan.SelectTo(pCell->render(), to);
  delete pCell;
}

/*
** This method will clear any existing text selection
*/
PUBLIC void ss::ClearTextSelection(){
  fastcan.Eval("select clear");
}

/*
** Return the current text selection.
** Return 0 if there is none.
** The return is malloc'ed; the caller must be careful to free it.
*/
PUBLIC char *ss::GetTextSelection(){
  fastcan.Eval("select item");
  const char *zId = evalStrResult();
  if( !zId || !*zId ) {
    return 0;
  }
  int id = atoi(zId);
  if( fastcan.Eval("index %d sel.first", id)!=TCL_OK ){
    return 0;
  }
  int from = evalIntResult();
  if( from < 0 ){
    return 0;
  }

  if( fastcan.Eval("index %d sel.last", id)!=TCL_OK ){
    return 0;
  }
  int to = evalIntResult();
  if( to < 0 || to < from ){
    return 0;
  }

  if( fastcan.Eval("itemcget %d -text", id)!=TCL_OK ){
    return 0;
  }
  const char *zText = evalStrResult();
  if( !zText || !*zText ){
    return 0;
  }
  int len = strlen(zText);
  if( from >= len ){
    return 0;
  }
  const char *zStart = zText + from;
  if( to >= len ){
    to = len-1;
  }
  int nChars = to - from + 1;
  char *zRange = new char[nChars + 1];
  strncpy(zRange, zStart, nChars);
  zRange[nChars] = 0;
  return zRange;
}

/*
** Scroll the canvas horizontally so that "fract"
** fraction of the total width of the canvas is off-screen to the left.
*/
PUBLIC void ss::XviewMoveto(float fract){
  if( fract < 0.0 ){
    fract = 0.0;
  }
  if( fract > 1.0 ){
    fract = 1.0;
  }

  /* Find out which column should be displayed leftmost;
  ** i.e which column is located the "fract" of the way 
  ** from the beginning
  */
  fastcan.Eval("cget -scrollregion");
  SsBbox bbox(evalStrResult());
  int leftPix = (int)((float) bbox.GetWidth() * fract);
  Xscroll(GetColAtDistance(1, leftPix));
}

/*
** Scroll the canvas horizontally iNum pages or units.
** If bUnits is nonzero, the scrolling is by unit; otherwise, it's by page.
*/
PUBLIC void ss::XviewScroll(
  int iNum,   // How many units to scroll
  int bUnits  // Nonzero if in units; otherwise, in pages
){
  if( bUnits ){
    int nCol = topleft.GetCol() + iNum;
    if( nCol < 1 ){
      nCol = 1;
    }
    if( nCol!=topleft.GetCol() ){
      Xscroll(nCol);
    }
  } else {
    int scrollableWidth = GetScrollableWidth();
    int nCol = GetColWithinDistance(topleft.GetCol(), scrollableWidth * iNum);
    if( nCol==topleft.GetCol() ){
      if( iNum<0 && nCol>1 ){
        nCol--;
      } else if( iNum>0 && nCol<sConfigure.cols-1 ){
        nCol++;
      }
    }
    if( nCol!=topleft.GetCol() ){
      Xscroll(nCol);
    }
  }
}

/*
** This method is called to scroll the spreadsheet horizontally so
** that "leftcol" is the leftmost visible column.
*/
PUBLIC void ss::Xscroll(int leftcol){

  /* Constrain the leftmost column to be within 1..MAX_LEFT,
  ** where MAX_LEFT is the highest numbered column that can be at the left.
  */
  if( leftcol < 1 ){
    leftcol = 1;
  }
  int scrollableWidth = GetScrollableWidth();
  int maxLeft = GetColWithinDistance(sConfigure.cols, -scrollableWidth);
  if( leftcol > maxLeft ){
    leftcol = maxLeft;
  }
  if( leftcol==topleft.GetCol() ) return;

  /* If there's a pending redisplay that will trim rows,
  ** just change the "topleft" member and return.
  ** The full redisplay will take care of positioning the spreadsheet.
  */
  if( (displayProperties & (DISPLAY_PENDING|DISPLAY_TRIM_ROWS))==
      (DISPLAY_PENDING|DISPLAY_TRIM_ROWS) ){
    topleft.SetCol(leftcol);
    return;
  }

  /* Adjust the view in the window so that "leftcol" is the
  ** non-title column displayed at the left of the window.
  ** Record the new top left cell.
  */
  int pix = GetDistanceBetweenCols(topleft.GetCol(), leftcol);
  fastcan.Move("mx", pix, 0);
  fastcan.Eval("xview scroll %d units", pix);
  topleft.SetCol(leftcol);
  DrawSelectionRectangle();
}

/*
** Scroll the canvas vertically so that "fract"
** fraction of the total height of the canvas is off-screen on top.
*/
PUBLIC void ss::YviewMoveto(float fract){
  if( fract < 0.0 ){
    fract = 0.0;
  }
  if( fract > 1.0 ){
    fract = 1.0;
  }

  /* Find out which row should be displayed topmost;
  ** i.e which column is located the "fract" of the way 
  ** from the beginning
  */
  fastcan.Eval("cget -scrollregion");
  SsBbox bbox(evalStrResult());
  int topPix = (int)((float) bbox.GetHeight() * fract);
  Yscroll(GetRowAtDistance(1, topPix), 0);
}

/*
** Scroll the canvas vertically iNum pages or units.
** If bUnits is nonzero, the scrolling is by unit; otherwise, it's by page.
*/
PUBLIC void ss::YviewScroll(
  int iNum,   // How many units to scroll
  int bUnits  // Nonzero if in units; otherwise, in pages
){
  if( bUnits ){
    int nRow = topleft.GetRow() + iNum;
    if( nRow < 1 ){
      nRow = 1;
    }
    if( nRow!=topleft.GetRow() ){
      Yscroll(nRow, 1);
    }
  } else {
    int scrollableHeight = GetScrollableHeight();
    int nRow = GetRowWithinDistance(topleft.GetRow(), scrollableHeight * iNum);
    if( nRow==topleft.GetRow() ){
      if( iNum<0 && nRow>1 ){
        nRow--;
      } else if( iNum>0 ){
        nRow++;
      }
    }
    if( nRow!=topleft.GetRow() ){
      Yscroll(nRow, 1);
    }
  }
}

/*
** This method is called to scroll the spreadsheet vertically so
** that "toprow" is the topmost visible row.
**
** If "bAddRows" is nonzero, then rows will be automatically added to
** fill in space at the bottom of the spreadsheet.
*/
PUBLIC void ss::Yscroll(int toprow, int bAddRows){
  if( nDisplayedRows==0 ) return;
  if( !bAddRows ){
    int height = CalculateViewportHeight();
    if( height>0 ){
      int scrollableHeight = height - GetConfiguredHeight(0);
      int lowestToprow = 
          GetRowWithinDistance(nDisplayedRows, -scrollableHeight);
      if( toprow>lowestToprow ){
        toprow = lowestToprow;
      }
    }
  }

  /* Constrain the toprow to be within 1..MAXROWS - 1.
  */
  if( toprow < 1 ){
    toprow = 1;
  }
  if( toprow >= sConfigure.maxRows ){
    toprow = sConfigure.maxRows - 1;
  }
  if( toprow==topleft.GetRow()) return;
  topleft.SetRow(toprow);
  Display(0, 1, bAddRows);
}

/*
** Return the next row after the given 'row' that does *not* contain
** displayable data.
*/
PUBLIC int ss::GetNextRowWithoutData(int row) {
  int nextRowWithoutData = row+1;
  if( !sConfigure.zVariable || !*sConfigure.zVariable ){
    return nextRowWithoutData;
  }
  for( ; nextRowWithoutData<sConfigure.maxRows; ++nextRowWithoutData ){
    int bRowHasData = 0;
    for( int col=1; col<sConfigure.cols; col++ ){
      SsCell cell(nextRowWithoutData, col);
      if( GetWidgetVariable(cell.render()) ){
        ++bRowHasData;
        break;
      }
    }
    if( !bRowHasData ){
      break;
    }
  }
  return nextRowWithoutData;
}

/*
** This method will create a new text glyph within the cell whose canvas id
** is given by "cellId".
** It returns the id of the glyph, or 0 if no glyph was created.
**
** This should only be called if the cell doesn't already have a text glyph.
*/
PUBLIC int ss::CreateTextGlyph(int cellId, const char *zString){
  if( !zString ) return 0;
  SsCell *pCell = fastcan.IdToCell(cellId);
  if( !pCell ) return 0;
  int width = GetColumnWidth(pCell->GetCol()) - cellPad*2;
  if( width<=0 ) return 0;
  char *zCell = pCell->render();

  if( fastcan.Eval("bbox %d", cellId)!=TCL_OK ) return 0;
  SsBbox bbox(evalStrResult());
  if( bbox.isnull() ) return 0;
  int halfRowHeight = GetConfiguredHeight(pCell->GetRow())/2;

  char *zTextFill = (char *)aFg[(void *)zCell];
  if( !zTextFill || !*zTextFill ){
    zTextFill = sConfigure.zFg;
  }
  int ret = fastcan.Eval("create text %d %d -text \"%q\" -width %d -fill %s "
        "-tags {notresize text} -anchor w -font \"%q\"",
        bbox.x0+2, bbox.y0+halfRowHeight, zString, width, zTextFill,
        sConfigure.zFont);
  int id = evalIntResult();
  assert(ret==TCL_OK && id>0);

  /* Make it so that the text is in the stacking order just above the cell.
  */
  fastcan.Eval("raise %d %d", id, cellId);
  fastcan.AddTag(zCell, id);
  fastcan.MapIdToCell(id, *pCell);

  /* If the newly created glyph is in the active cell, tag it as such.
  */
  if( pCell->equal(activeCell) && fastcan.FindFirst("active")==0 ){
    fastcan.AddTag("active", zCell);
  }
  return id;
}

/*
** Utilities
*/

/*
** This procedure will fire whenever the global variable associated with
** the spreadsheet widget is either modified or unset.
*/
PUBLIC STATIC char *
ss::SsVarProc(
  ClientData clientData,  // a pointer to the spreadsheet object
  Tcl_Interp *interp,
  const char *zName,      // the name of the global variable
  const char *zIndex,     // the name of the changed element in the variable
  int flags               // whether the variable was written or unset
){

  ss *p = (ss *)clientData;
  if( p->fromSs ) return 0;

  p->fromSs = 1;  // Avoids looping on widget variable modifications
  p->MonitorVar(zName, zIndex, flags);
  p->fromSs = 0;
  return 0;
}

/*
** This is a member function called by SsVarProc() that keeps track of
** changes to the widget variable.
**
** The single cell associated with the modified or deleted widget variable
** entry will be redrawn.  But if the spreadsheet is already scheduled
** for redisplay, we'll wait for the redisplay.
*/
PUBLIC void ss::MonitorVar(const char *zName, const char *zElement, int flags){
  SsCell cell;

  /* Determine which cell was modified.
  ** If the modified cell is the currently active cell,
  ** copy its value to the "active" entry.
  ** This must be done before checking if a full redisplay is pending below.
  */
  if( flags & TCL_TRACE_WRITES ){
    if( zElement ){
      if( *zElement=='a' && strcmp(zElement, "active")==0 ){
        cell.set(activeCell);
      } else {
        bDirtyCell++;
        cell.set(zElement);
        if( !cell.isnull() && cell.equal(activeCell) ){

          /* Only set the "active" entry if it's really changing;
          ** this avoids firing additional traces
          */
          const char *zOldActive = GetWidgetVariable("active");
          const char *zOldElt = GetWidgetVariable(zElement);
          if( !zOldActive || !zOldElt || strcmp(zOldActive, zOldElt)!=0 ){
            SetWidgetVariable("active", GetWidgetVariable(zElement));
          }
        }
      }
    }
  } else if( flags & TCL_TRACE_UNSETS ){
    bDirtyCell++;
    if( !zElement ){  // the whole array is gone.  Delete all cell text
      fastcan.Eval("itemconfigure text -text {}");
    } else {
      if( *zElement=='a' && strcmp(zElement, "active")==0 ){
        cell.set(activeCell);
      } else {
        cell.set(zElement);
        if( !cell.isnull() && cell.equal(activeCell) ){

          /* Only set the "active" entry if it's really changing;
          ** this avoids firing additional traces
          */
          const char *zOldActive = GetWidgetVariable("active");
          if( zOldActive ){
            SetWidgetVariable("active", 0);
          }
        }
      }
    }
  } else {
    return;
  }

  /* If a full redisplay with row trimming is pending, 
  ** we need not do anything else.
  ** The full redisplay will take care of everything.
  */
  if( (displayProperties & (DISPLAY_PENDING|DISPLAY_TRIM_ROWS))==
      (DISPLAY_PENDING|DISPLAY_TRIM_ROWS) ){
    return;
  }

  /* If the cell change is out of bounds of the configured maximums, ignore it.
  **
  ** If the cell is outside the currently displayed spreadsheet,
  ** schedule the spreadsheet to be redisplayed and its number of 
  ** rows trimmed.  (This will cause any future trace firings to
  ** be ignored until the spreadsheet is redrawn.)
  **
  ** Otherwise, recalculate the cell's text and the height of its row,
  ** if the spreadsheet isn't already scheduled for redisplay.
  */
  if( !cell.isnull() ){
    if( cell.GetRow() < 0 || cell.GetCol() < 0 || 
        cell.GetRow() > sConfigure.maxRows || 
        cell.GetCol() > sConfigure.cols ){
      // ignore the cell change
    } else if( cell.GetRow() > 0 && cell.GetCol() > 0 && 
        cell.GetRow() >= nDisplayedRows ){
      Display(0, /* trim == */ 1);
    } else if( !(displayProperties & DISPLAY_PENDING) ){
      const char *zValue = 0;
      if( zElement && (flags & TCL_TRACE_WRITES) ){
        zValue = GetWidgetVariable(zElement);
      }
      fastcan.SetText(cell.render(), zValue, this);
      if( cell.GetRow() > 0 && cell.GetCol() > 0 && zValue && *zValue ){
        ResetHeight(cell);
      }
    }
  }
}

/*
** Set the widget variable's element called "zName" to the value
** pointed to by "zValue".  If there is no widget variable, nothing happens.
*/
PRIVATE void ss::SetWidgetVariable(char *zName, const char *zValue){
  if( sConfigure.zVariable && *sConfigure.zVariable ){
    Tcl_SetVar2(interp, sConfigure.zVariable, zName, zValue, TCL_GLOBAL_ONLY);
  }
}

/*
** Unset the entry in the widget variable named zName.
*/
PRIVATE void ss::UnsetWidgetVariable(char *zName){
  if( sConfigure.zVariable && *sConfigure.zVariable ){
    Tcl_UnsetVar2(interp, sConfigure.zVariable, zName, TCL_GLOBAL_ONLY);
  }
}

/*
** Given the widget variable's element called "zName", return its value.
** Return 0 if there is no value, or if there is no widget variable.
*/
PRIVATE const char *ss::GetWidgetVariable(const char *zName){
  if( sConfigure.zVariable && *sConfigure.zVariable ){
    return Tcl_GetVar2(interp, sConfigure.zVariable, zName, TCL_GLOBAL_ONLY);
  } else {
    return 0;
  }
}

/*
** Return a pointer to the currently active cell.
** Returns 0 if there is no active cell.
*/
PRIVATE SsCell *ss::GetActiveCell(){
  if( activeCell.isnull() ){
    return 0;
  } else {
    return &activeCell;
  }
}

/*
** Set the "active" entry in the widget variable.
*/
PRIVATE void ss::SetActiveVar(){
  fromSs = 1;  // turns off tracing
  SsString *pValue = fastcan.Itemcget("active", "-text");
  SetWidgetVariable("active", pValue->render());
  delete pValue;
  fromSs = 0;
}

/*
** Commit the value currently in the active cell by writing the value
** into the widget's designated variable entry for that cell.
*/
PRIVATE void ss::CommitActiveValue(){
  if( !sConfigure.zVariable || !*sConfigure.zVariable ) return;

  if( activeCell.isnull() ) return;
  char *zCell = activeCell.render();

  const char *zCellValue = GetWidgetVariable("active");

  /* Avoid unnecessary trace firings by checking to see if the active value
  ** matches the cell value first.
  */
  const char *zVarValue = GetWidgetVariable(zCell);
  if( zVarValue && zCellValue && strcmp(zCellValue, zVarValue)==0 ) {
    return;
  }

  /* Note that we always attempt to minimize the size of the widget variable
  ** by unsetting its cell entry if the widget has no data in that cell.
  **
  ** Also note that we allow the MonitorVar() trace on the cell to fire, 
  ** even though in most cases it will do nothing.  There
  ** is a possibility that a validating application trace might change 
  ** the value of the cell, and leaving MonitorVar() on will catch that case.
  */
  if( zCellValue && *zCellValue ){
    SetWidgetVariable(zCell, zCellValue);
  } else {
    UnsetWidgetVariable(zCell);
  }
}

/*
** Configure the widget object.
** We can't use the Tk version of this function (Tk_ConfigureWidget) since
** the ss widget isn't actually a tkwin.
*/
PUBLIC void SsConfigure::ConfigureWidgetDefaults(){

  /* Make both the argvName (e.g. "-activebackground") and the dbName
  ** (e.g. "activeBackground") refer to the same configuration specification.
  ** This will allow options to be looked up by synonym.
  */
  const Tk_ConfigSpec *pSsSpec = ssSpecs;
  while( pSsSpec->type!=TK_CONFIG_END ){
    char *addr;
    switch( pSsSpec->type ){
      case TK_CONFIG_STRING:
        addr = (char *)this;
        addr += pSsSpec->offset;
        ChangeStr((char **)addr, pSsSpec->defValue);
        aOptions.insert(pSsSpec->argvName, (void*)pSsSpec);
        aOptions.insert((void*)pSsSpec->dbName,  (void*)pSsSpec);
        break;

      case TK_CONFIG_INT:
        addr = (char *)this;
        addr += pSsSpec->offset;
        *(int *)addr = atoi(pSsSpec->defValue);
        aOptions.insert(pSsSpec->argvName,  (void*)pSsSpec);
        aOptions.insert((void*)pSsSpec->dbName,  (void*)pSsSpec);
        break;

      case TK_CONFIG_SYNONYM:
        aOptions.insert(pSsSpec->argvName, (void *)pSsSpec);
        break;
    }
    pSsSpec++;
  }
}

/*
** Return a list that contains a sublist for each configuration item.
** The list is malloc'ed; the caller must be careful to free it.
*/
PUBLIC SsTclList *SsConfigure::GetConfig(){
  SsTclList *pConfig = new SsTclList(0);
  const Tk_ConfigSpec *pSsSpec = ssSpecs;
  while( pSsSpec->type!=TK_CONFIG_END ){
    SsTclList entry(0);
    entry.append(pSsSpec->argvName);
    char *addr;
    switch( pSsSpec->type ){
      case TK_CONFIG_STRING:
      case TK_CONFIG_INT:
        entry.append(pSsSpec->dbName);
        entry.append(pSsSpec->dbClass);
        entry.append(pSsSpec->defValue);
        entry.append(GetOptionValue(pSsSpec->argvName));
        break;

      case TK_CONFIG_SYNONYM:
        entry.append(pSsSpec->dbName);
        break;
    }
    pConfig->append(&entry);
    pSsSpec++;
  }
  return pConfig;
}

/*
** Return nonzero if zOption is a valid spreadsheet option
*/
PUBLIC int SsConfigure::IsOption(const char *zOption){
  return (int) aOptions[(void *)zOption];
}

/*
** Return the value of the given option.
** Returns 0 if there is no such option.
** zOption may either be an "argv name" (e.g. "-activebackground")
** or a "db name" (e.g. "activeBackground").
**
** The return may be a static value.
*/
PUBLIC char *SsConfigure::GetOptionValue(const char *zOption){
  static char zInt[33];  // guaranteed to fit any 32-bit integer
  Tk_ConfigSpec *pSsSpec = (Tk_ConfigSpec *)aOptions[(void *)zOption];
  if( pSsSpec==0 ){
    return 0;
  }
  char *addr = (char *)this;
  switch( pSsSpec->type ){
    case TK_CONFIG_STRING:
      addr += pSsSpec->offset;
      return *(char **)addr;
    case TK_CONFIG_INT:
      addr += pSsSpec->offset;
      sprintf(zInt, "%d", *(int *)addr);
      return zInt;
    case TK_CONFIG_SYNONYM:
      return GetOptionValue(pSsSpec->dbName);
    default:
      return 0;
  }
}

/*
** Set the value of the given option to the given value.
** If zOption isn't a valid option for the spreadsheet, nothing happens.
** zOption may either be an "argv name" (e.g. "-activebackground")
** or a "db name" (e.g. "activeBackground").
*/
PUBLIC void SsConfigure::SetOptionValue(const char *zOption, 
    const char *zValue){
  Tk_ConfigSpec *pSsSpec = (Tk_ConfigSpec *)aOptions[(void *)zOption];
  if( pSsSpec==0 ){
    return;
  }
  char *addr = (char *)this;
  addr += pSsSpec->offset;
  switch( pSsSpec->type ){
    case TK_CONFIG_STRING:
      ChangeStr((char **)addr, zValue);
      break;
    case TK_CONFIG_INT:
      *(int *)addr = atoi(zValue);
      break;
    case TK_CONFIG_SYNONYM:
      SetOptionValue(pSsSpec->dbName, zValue);
      break;
  }
}

/*
** Return the text selection's id in pId, and its range in
** pFrom and pTo.  Return nonzero if successful.
** The return values are meaningless if the return value is 0.
*/
PRIVATE int ss::FindTextSelection(int *pId, int *pFrom, int *pTo){
  fastcan.Eval("select item");
  *pId = evalIntResult();
  if( *pId==0 ){
    return 0;
  }
  if( fastcan.Eval("index %d sel.first", *pId)!=TCL_OK ){
    return 0;
  }
  *pFrom = evalIntResult();
  if( fastcan.Eval("index %d sel.last", *pId)!=TCL_OK ){
    return 0;
  }
  *pTo = evalIntResult();
  return 1;
}

/*
** Utility routines.
*/

/*
** Eval the results of a string.
*/
PUBLIC int ss::evalf(const char *zFormat, ...){
  va_list ap;
  int result;
  va_start(ap,zFormat);
  result = ::evalf(interp, zFormat, ap);
  va_end(ap);
  return result;
}

/*
** Return the string result of the most recent Tcl command.
*/
PUBLIC const char *ss::evalStrResult(){
  return ::evalStrResult(interp);
}

/*
** Return the integer result of the most recent Tcl command.
*/
PUBLIC int ss::evalIntResult(){
  return ::evalIntResult(interp);
}

/*
** Print a debugging message
*/
PUBLIC void ss::Trace(int mask, const char *zFormat, ...){
  va_list ap;
  va_start(ap, zFormat);
  ::Trace(interp, mask, zFormat, ap);
  va_end(ap);
}

#ifndef WIN32
/*
** This proc will record the current time, with microsecond granularity
** (where available).
** If bPrint is true, then it will print the time since the last click.
**
** Used for performance testing.
*/
static struct timeval sTv;
PRIVATE int ss::Click(int bPrint, char *zMsg){
  struct timeval tv;
  struct timezone tz;
  gettimeofday(&tv, &tz);
  int delta = (tv.tv_sec - sTv.tv_sec)*1000000 + (tv.tv_usec - sTv.tv_usec);
  if( bPrint ){
    fprintf(stderr, "%s: %d clicks\n", zMsg, delta);
  }
  sTv = tv;
  return delta;
}
PRIVATE int ss::Click(){
  return Click(0, 0);
}
#endif
