/*
** This module implements a class encapsulating a Tcl list.
*/
static char const rcsid[] = "@(#) $Header: /home/softools/cvs/softools/code/common/cked/src/ss_list.cpp,v 1.5 2003/01/07 20:05:59 mark Exp $";

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <tcl.h>
#include "ss_list.h"

#if INTERFACE
/*
** object: SsTclList
**
** A single instance of the SsTclList class represents a list implemented
** by a Tcl list.  Methods to append, index, iterate on,
** and execute the list are provided.
**
** To iterate on the list, use the following idiom:
**  for( char *z=TclList.first(); z; z=TclList.next() ){
**    DO SOMETHING;
**  }
*/
class SsTclList {
private:
  Tcl_Obj *pList;      // the Tcl list being managed
  int lastElement;     // Index of the last Tcl element visited
  int nobjc;           // Number of elements in the list, after cache()
  Tcl_Obj **ppObjv;    // Pointer to the elements in the list, after cache()

  void invalidate() {ppObjv = 0; nobjc = 0;}
  int cache(){
    if( !ppObjv ){
      if( Tcl_ListObjGetElements(interp, pList, &nobjc, &ppObjv)!=TCL_OK ){
        return 0;
      }
    }
    return 1;
  }

protected:
  Tcl_Interp *interp;  // Tcl interpreter for this list
public:
  SsTclList(Tcl_Interp *pInterp) {init(pInterp);}
  SsTclList(Tcl_Interp *pInterp, const char *zValue){
    init(pInterp);
    set(zValue);
  }
  SsTclList(Tcl_Interp *pInterp, Tcl_Obj *pObj){
    init(pInterp);
    set(pObj);
  }
  ~SsTclList(){
    Tcl_DecrRefCount(pList);
  }

  int Execute(){ return Tcl_EvalObjEx(interp, pList, TCL_EVAL_DIRECT); }
  Tcl_Obj *get_listrep(void) const {return pList;}
  int count(void) {cache(); return nobjc;}

  char *render() const { return Tcl_GetString(pList); }

  /*
  ** This method sets the value of the list in pList to the string in zStr.
  ** If zStr isn't the string representation of a valid Tcl list, 
  ** then bad things could happen.
  */
  void set(const char *zStr) {Tcl_SetStringObj(pList, zStr, -1); invalidate();}

  void append(const char *zStr){
    Tcl_ListObjAppendElement(interp, pList, Tcl_NewStringObj(zStr, -1));
    invalidate();
  }
  void append(int num){
    Tcl_ListObjAppendElement(interp, pList, Tcl_NewIntObj(num));
    invalidate();
  }

  /*
  ** Concatenate the elements on the SsTclList referenced by pList, 
  ** to this list.
  */
  void concat(SsTclList *pNew){
    Tcl_ListObjAppendList(interp, pList, pNew->pList);
    invalidate();
  }

  /*
  ** Append the tcl list pNew to the current list.
  ** This creates a single new element in this list.  Use concat() if
  ** each element from pNew should create a new element in the current list.
  */
  void append(SsTclList *pNew){
    pNew->cache();
    Tcl_ListObjAppendElement(interp, pList, 
        Tcl_NewListObj(pNew->nobjc, pNew->ppObjv));
    invalidate();
  }
  char *first() {return start() ? next() : 0;}
  char *next(){
    return ( ppObjv && nobjc>lastElement+1 ) ?
        Tcl_GetString(ppObjv[++lastElement]) : 0;
  }

  /*
  ** Set the i-th entry on the list to an integer or a string.
  ** When doing this, the list's string representation, if any, must be freed.
  */
  void lset(int i, int v){
    if( cache() && i>=0 && i<nobjc ){
      Tcl_SetIntObj(ppObjv[i], v);
      Tcl_InvalidateStringRep(pList);
    }
  }
  void lset(int i, const char *z){
    if( cache() && i>=0 && i<nobjc ){
      Tcl_SetStringObj(ppObjv[i], z, -1);
      Tcl_InvalidateStringRep(pList);
    }
  }

  /*
  ** Return the first list element on the list as a Tcl_Obj pointer.
  ** Return 0 if there are no elements on the list.
  */
  Tcl_Obj *first_list() {return start() ? next_list() : 0;}

  /*
  ** Return the next list element on the list as a Tcl_Obj pointer.
  ** Return 0 if there are no more.
  */
  Tcl_Obj *next_list(){
    return( ppObjv && nobjc>lastElement+1 ) ?  ppObjv[++lastElement] : 0;
  }

  int start(){ lastElement = -1; return cache(); }

  Tcl_Obj *lindex(int i) {return cache() ? ppObjv[i] : 0;}
  int isnull() {return cache() ? (nobjc==0) : 1;}
}
#endif

/*
** Initialize the list
*/
PRIVATE void SsTclList::init(Tcl_Interp *pInterp){
  interp = pInterp;
  pList = Tcl_NewListObj(0, 0);
  nobjc = 0;
  ppObjv = 0;
  lastElement = -1;
  Tcl_IncrRefCount(pList);
}

/*
** This method sets the value of the list in pList to the Tcl object in pObj.
*/
PUBLIC void SsTclList::set(Tcl_Obj *pObj){
  if( !pObj ) return;
  int objc;
  Tcl_Obj **objv;
  int ret = Tcl_ListObjGetElements(interp, pObj, &objc, &objv);
  if( ret==TCL_OK ){
    Tcl_SetListObj(pList, objc, objv);
    invalidate();
  }
}

/*
** Return the first integer element on the list in the return parameter
** pInt.  The method itself returns 0 if there are no elements in the list.
*/
PUBLIC int SsTclList::first_int(int *pInt){
  return start() ? next_int(pInt) : 0;
}

/*
** Return the next integer element on the list in the return parameter pInt.
** The method itself returns 0 if there are no more elements in the list.
*/
PUBLIC int SsTclList::next_int(int *pInt){
  assert(ppObjv!=0);
  lastElement++;
  if( nobjc>lastElement ){
    Tcl_GetIntFromObj(interp, ppObjv[lastElement], pInt);
    return 1;
  } else {
    lastElement = -1;
    return 0;
  }
}

/*
** Return the i-th entry in the list as a character string.
** 0 is returned on error.
*/
PUBLIC char *SsTclList::get(int i){
  if( !cache() ){
    return 0;
  }
  if( i>=0 && i<nobjc ){
    return Tcl_GetString(ppObjv[i]);
  } else {
    return 0;
  }
}

/*
** Return the i-th entry in the list as an integer.
** -1 is returned on error.
*/
PUBLIC int SsTclList::get_int(int i){
  if( !cache() ){
    return -1;
  }
  int n = -1;
  if( i>=0 && i<nobjc ){
    Tcl_GetIntFromObj(interp, ppObjv[i], &n);
  }
  return n;
}
