/*
** This file implements the Tcl interface to the "ss" widget
*/
static char const rcsid[] = "@(#) $Header: /home/softools/cvs/softools/code/common/cked/src/ss_tcl.cpp,v 1.41 2004/06/15 18:36:31 mark Exp $";

#include <tcl.h>
#include <tk.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include "ss_tcl.h"

/*
** Called when the TCL command associated with a ss is deleted.
*/
PUBLIC STATIC void ss::DeleteProc(void *pArg){
  ss *p = (ss*)pArg;
  delete p;
}

/*
** When the TCL command associated with a ss is executed, this
** procedure runs to implement the command.
*/
PUBLIC STATIC int ss::CmdProc(
  ClientData ClientData,   // A pointer to an ss
  Tcl_Interp *interp,      // The interpreter controlling the ss
  int objc,                // Number of arguments
  Tcl_Obj *const* objv     // Value of each argument
){
  static const char *SS_strs[] = {
    "activate",         "bbox",           "canvas",
    "cellconfigure",    "cget",           "clone",
    "color",            "configure",      "curselection",
    "curtextsel",       "curvalue",       "delete",
    "displayedrows",    "fast_canvas",    "height",
    "hide",             "icursor",        "index",
    "insert",           "minheight",      "minwidth",
    "nextemptyrow",     "redisplay",      "see",
    "selection",        "selrect",        "set",
    "textselection",    "trim",           "viewport",
    "width",            "xview",          "yview",
    0                    
  };
  enum SS_enum {
    SS_ACTIVATE,        SS_BBOX,            SS_CANVAS,
    SS_CELLCONFIGURE,   SS_CGET,            SS_CLONE,
    SS_COLOR,           SS_CONFIGURE,       SS_CURSELECTION,
    SS_CURTEXTSEL,      SS_CURVALUE,        SS_DELETE,
    SS_DISPLAYEDROWS,   SS_FAST_CANVAS,     SS_HEIGHT,
    SS_HIDE,            SS_ICURSOR,         SS_INDEX,
    SS_INSERT,          SS_MINHEIGHT,       SS_MINWIDTH,
    SS_NEXTEMPTYROW,    SS_REDISPLAY,       SS_SEE,
    SS_SELECTION,       SS_SELRECT,         SS_SET,
    SS_TEXTSELECTION,   SS_TRIM,            SS_VIEWPORT,
    SS_WIDTH,           SS_XVIEW,           SS_YVIEW,
  };
  ss *p = (ss*)ClientData;
  int index;

  if( objc<2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "METHOD ?ARG ...?");
    return TCL_ERROR;
  }

  p->Trace(TRACE_COMMAND, "WIDGET: %s %s %s %s %s\n", 
      Tcl_GetStringFromObj(objv[0], 0), Tcl_GetStringFromObj(objv[1], 0),
      (objc>2?Tcl_GetStringFromObj(objv[2],0):""),
      (objc>3?Tcl_GetStringFromObj(objv[3],0):""),
      (objc>4?Tcl_GetStringFromObj(objv[4],0):"")
      );

  if( Tcl_GetIndexFromObj(interp, objv[1], SS_strs, "option", 0, &index)){
    return TCL_ERROR;
  }

  switch( (enum SS_enum) index ){

    /*
    ** synopsis: SS activate INDEX
    ** Make the cell with the given INDEX the active cell.
    */
    case SS_ACTIVATE:
    {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "INDEX");
        return TCL_ERROR;
      }
      char *zIndex = Tcl_GetStringFromObj(objv[2], 0);
      p->Activate(zIndex);
      break;
    }

    /*
    ** synopsis:  SS bbox FIRST ?LAST?
    ** A bounding box consisting of a list
    ** of the form {x0, y0, x1, y1} is returned.
    ** The bounding box is in canvas coordinates.
    */
    case SS_BBOX:
    {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "FIRST ?LAST?");
        return TCL_ERROR;
      }
      char *zFirst = Tcl_GetStringFromObj(objv[2], 0);
      char *zLast;
      if( objc==3 ){
        zLast = zFirst;
      } else {
        zLast = Tcl_GetStringFromObj(objv[3], 0);
      }
      if( !*zFirst || !*zLast ){
        Tcl_ResetResult(interp);
        return TCL_OK;
      }
      SsBbox *pBbox = p->Bbox(zFirst, zLast);
      if( pBbox ){
        SsTclList result(interp);
        result.append(pBbox->x0);
        result.append(pBbox->y0);
        result.append(pBbox->x1);
        result.append(pBbox->y1);

        Tcl_SetObjResult(interp, result.get_listrep());
        delete pBbox;
      } else {
        Tcl_ResetResult(interp);
      }
      break;
    }

    /*
    ** synopsis: SS canvas SCRIPT
    ** Evaluate the SCRIPT as a canvas command on the underlying Tk canvas.
    */
    case SS_CANVAS:
    {
      if( objc<3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "SCRIPT");
        return TCL_ERROR;
      }
      SsTclList cmd(interp);
      cmd.append(p->fastcan.zName);
      for( int i=2; i<objc; ++i ){
        cmd.append(Tcl_GetStringFromObj(objv[i], 0));
      }
      return cmd.Execute();
    }

    /*
    ** synopsis:  SS cellconfigure CELL ?OPTION VALUE? ?OPTION VALUE?...
    **            SS cellconfigure default ?OPTION?
    **            SS cellconfigure fg COLOR CELL ?CELL?...
    **            SS cellconfigure bg COLOR CELL ?CELL?...
    **
    ** "SS cellconfigure CELL ..." configures the options for a particular cell.
    ** Currently, only foreground and background colors can be configured.
    **
    ** "SS cellconfigure default", with no arguments, sets all the options
    ** to their default values, for all cells.
    ** "SS cellconfigure default OPTION" sets just
    ** the given option to its default value.
    **
    ** "SS cellconfigure fg" and "SS cellconfigure bg" are used to color
    ** the foreground or background color of a group of cells.
    */
    case SS_CELLCONFIGURE:
    {
      if( objc<3 ){
        Tcl_WrongNumArgs(interp, 2, objv, 
            "CELL ?OPTION VALUE? ?OPTION VALUE?...\ndefault ?OPTION?"
            "\nfg COLOR CELL ?CELL?...\nbg COLOR CELL ?CELL?...");
        return TCL_ERROR;
      }
      char *zArg = Tcl_GetStringFromObj(objv[2], 0);
      if( *zArg=='d' && strcmp(zArg, "default")==0 ){
        if( objc>3 ){
          p->CellConfigureDefault(Tcl_GetStringFromObj(objv[3], 0));
        } else {
          p->CellConfigureDefault(0);
        }
      } else if( *zArg=='f' && strcmp(zArg, "fg")==0 ){
        if( objc<5 ){
          Tcl_WrongNumArgs(interp, 2, objv, "fg COLOR CELL ?CELL?...");
          return TCL_ERROR;
        }
        p->CellConfigureFg(Tcl_GetStringFromObj(objv[3], 0), objc-4, objv+4);
      } else if( *zArg=='b' && strcmp(zArg, "bg")==0 ){
        if( objc<5 ){
          Tcl_WrongNumArgs(interp, 2, objv, "bg COLOR CELL ?CELL?...");
          return TCL_ERROR;
        }
        p->CellConfigureBg(Tcl_GetStringFromObj(objv[3], 0), objc-4, objv+4);
      } else {
        char *zCell = zArg;

        if( objc==3 ){
          char *zFg = (char *)(p->aFg[(void *)zCell]);
          if( !zFg ){
            zFg = "{}";
          }
          char *zBg = (char *)(p->aBg[(void *)zCell]);
          if( !zBg ){
            zBg = "{}";
          }
          Tcl_AppendResult(interp, "-fg ", zFg, " -bg ", zBg, 0);
        } else {
          if( (objc%2)!=1 ){
            Tcl_AppendResult(interp, "Bad option specifier for cellconfigure",
                0);
            return TCL_ERROR;
          }
          p->CellConfigure(zCell, objc-3, objv+3);
        }
      }
      break;
    }

    /*
    ** synopsis:  SS cget -OPTION
    ** Return the value of the configured OPTION.
    */
    case SS_CGET:
    {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "-OPTION");
        return TCL_ERROR;
      }
      char *zOption = Tcl_GetStringFromObj(objv[2], 0);
      char *zValue = p->Cget(zOption);
      Tcl_ResetResult(interp);
      if( zValue ){
        Tcl_AppendResult(interp, zValue, 0);
      } else {
        Tcl_AppendResult(interp, "Bad option ", zOption, 0);
        return TCL_ERROR;
      }
      break;
    }

    /*
    ** synopsis:  SS clone FROM TO CANVAS ROW_MAG
    ** "Clone" the part of the spreadsheet within the range of rows
    ** given by FROM and TO, and write the spreadsheet on CANVAS.
    ** ROW_MAG is the magnification factor for each row.
    */
    case SS_CLONE:
    {
      if( objc!=6 ){
        Tcl_WrongNumArgs(interp, 2, objv, "FROM TO CANVAS ROW_MAG");
        return TCL_ERROR;
      }
      int from = atoi(Tcl_GetStringFromObj(objv[2], 0));
      int to = atoi(Tcl_GetStringFromObj(objv[3], 0));
      char *zNewCanvas = Tcl_GetStringFromObj(objv[4], 0);
      char *zRowMag = Tcl_GetStringFromObj(objv[5], 0);
      float rowMag;
      if( sscanf(zRowMag, "%f", &rowMag)!=1 ){
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "ROW_MAG is not a floating point value", 0);
        return TCL_ERROR;
      }

      p->evalf("winfo exists %s", zNewCanvas);
      if( !p->evalIntResult() ){
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "Canvas does not exist", 0);
        return TCL_ERROR;
      }
      p->Clone(from, to, zNewCanvas, rowMag);
      Tcl_ResetResult(interp);
      break;
    }

    /*
    ** synopsis: SS color CELL
    ** Return the foreground and background colors of CELL as a list.
    */
    case SS_COLOR:
    {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "CELL");
        return TCL_ERROR;
      }
      SsCell cell(Tcl_GetStringFromObj(objv[2], 0));
      Tcl_AppendResult(interp, p->GetCellFg(cell), " ", p->GetCellBg(cell), 0);
      break;
    }

    /*
    ** synopsis: SS configure ?OPTION ?VALUE? ?OPTION VALUE?...
    ** With no arguments, return the complete list of OPTION VALUE pairs
    ** as a list.
    ** With a single OPTION argument, return the value of the configured OPTION.
    ** With more arguments, set each specified OPTION's configured value
    ** to the given VALUE.
    */
    case SS_CONFIGURE:
      if( objc==2 ){
        SsTclList *pList = p->sConfigure.GetConfig();
        Tcl_AppendResult(interp, pList->render(), 0);
        delete pList;
        return TCL_OK;
      }
      if( objc<4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?OPTION VALUE?...");
        return TCL_ERROR;
      }
      if( (objc%2)!=0 ){
        Tcl_AppendResult(interp, "Need even # arguments for configure", 0);
        return TCL_ERROR;
      }
      for( int i=2; i<objc; i+=2 ){
        char *zOption = Tcl_GetStringFromObj(objv[i], 0);
        if( !p->sConfigure.IsOption(zOption) ){
          Tcl_ResetResult(interp);
          Tcl_AppendResult(interp, zOption, " not a valid option", 0);
          return TCL_ERROR;
        }
      }
      return p->Configure(objc-2, objv+2);

    /*
    ** synopsis: SS curselection
    ** Return the sorted list of cells that are currently selected.
    */
    case SS_CURSELECTION:
    {
      if( objc!=2 ){
        Tcl_WrongNumArgs(interp, 2, objv, "");
        return TCL_ERROR;
      }
      SsTclList *pSelection = p->GetCurrentSelection();
      Tcl_ResetResult(interp);
      if( pSelection ){
        Tcl_SetObjResult(interp, pSelection->get_listrep());
        delete pSelection;
      }
      break;
    }

    /*
    ** synopsis: SS curtextsel
    ** Returns: a list of the form "CELL FROM TO",
    ** where CELL is the cell with the text selection,
    ** and FROM and TO are the indices in that cell defining the range
    ** of the selection.
    */
    case SS_CURTEXTSEL:
    {
      if( objc!=2 ){
        Tcl_WrongNumArgs(interp, 2, objv, "");
        return TCL_ERROR;
      }
      SsTclList *pTextSelection = p->GetCurrentTextSelection();
      Tcl_ResetResult(interp);
      if( pTextSelection ) {
        Tcl_SetObjResult(interp, pTextSelection->get_listrep());
        delete pTextSelection;
      }
      break;
    }

    /*
    ** synopsis:  SS curvalue ?VALUE?
    ** With no arguments, returns the value in the active cell.
    ** If VALUE is specified, it sets the value in the active cell to VALUE.
    */
    case SS_CURVALUE:
      if( objc<2 || objc>3){
        Tcl_WrongNumArgs(interp, 2, objv, "curvalue ?VALUE?");
        return TCL_ERROR;
      }
      if( objc==2 ){
        SsString *pCurrentValue = p->GetCurrentValue();
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, pCurrentValue->render(), 0);
        delete pCurrentValue;
      } else {
        p->SetCurrentValue(Tcl_GetStringFromObj(objv[2], 0));
      }
      break;

    /*
    ** synopsis:  SS delete INDEX FROM ?TO?
    ** Delete characters from the cell specified by INDEX, from the
    ** FROM-th character, to the TO-th character.  If TO is not specified,
    ** just delete the single FROM-th character.
    */
    case SS_DELETE:
    {
      if( objc<4 || objc>5 ){
        Tcl_WrongNumArgs(interp, 2, objv, "INDEX FROM ?TO?");
        return TCL_ERROR;
      }
      char *zIndex = Tcl_GetStringFromObj(objv[2], 0);
      SsCell cell;
      if( strcmp(zIndex, "active")==0 ){
        cell.set(p->GetActiveCell());
      } else {
        cell.set(zIndex);
        if( cell.isnull() ){
          Tcl_AppendResult(interp, "Bad format for index ", zIndex, 0);
          return TCL_ERROR;
        }
      }
      char *zFrom = Tcl_GetStringFromObj(objv[3], 0);
      char *zTo = zFrom;
      if( objc==5 ){
        zTo = Tcl_GetStringFromObj(objv[4], 0);
      }
      p->Delete(cell, zFrom, zTo);
      break;
    }

    /*
    ** synopsis:  SS displayedrows
    ** Return the number of rows that are currently being displayed.
    */
    case SS_DISPLAYEDROWS:
      if( objc!=2 ){
        Tcl_WrongNumArgs(interp, 2, objv, 
            "height default\nheight ?ROW? ?VALUE? ?ROW VALUE?...");
        return TCL_ERROR;
      }
      char zInt[33];
      sprintf(zInt, "%d", p->nDisplayedRows);
      Tcl_AppendResult(interp, zInt, 0);
      break;

    /*
    ** synopsis: SS fast_canvas CMD ARGS...
    ** where CMD is the command to pass to the underlying canvas
    ** and ARGS are the arguments to that canvas command.
    ** The command is called "fast_canvas" because it makes use
    ** of the spreadsheet's special implementation of canvas tags,
    ** which looks up canvas id's in constant rather than linear time.
    */
    case SS_FAST_CANVAS:
      return p->FastCanvasCmd(objc, objv);
      break;

    /*
    ** synopsis:  SS height ?ROW? ?VALUE? ?ROW VALUE?...
    **            SS height default
    ** With no arguments, returns a list of rows and their heights in pixels.
    ** With one ROW argument, returns its height in pixels.
    ** If at least one ROW VALUE pair is specified, it sets each 
    ** specified ROW's height to the specified VALUE.
    */
    case SS_HEIGHT:
    {
      if( objc<2 ){
        Tcl_WrongNumArgs(interp, 2, objv, 
            "height default\nheight ?ROW? ?VALUE? ?ROW VALUE?...");
        return TCL_ERROR;
      }

      if( objc==2 ){

        /* Return the heights of all configured rows as an even-numbered
        ** list of rows and heights
        */
        SsTclList *pHeights = p->GetAllHeights();
        Tcl_SetObjResult(interp, pHeights->get_listrep());
        delete pHeights;
        return TCL_OK;
      }

      if( objc==3 ){
        char *zArg = Tcl_GetStringFromObj(objv[2], 0);
        if( strcmp(zArg, "default")==0 ){  // reset all heights to defaults
          p->ResetConfiguredHeights();
        } else {
          int row;
          if( sscanf(zArg, "%d", &row)!=1 ){
            Tcl_AppendResult(interp, "Argument ", zArg, " not an integer", 0);
            return TCL_ERROR;
          }
          SsString res(p->GetConfiguredHeight(row));
          Tcl_ResetResult(interp);
          Tcl_AppendResult(interp, res.render(), 0);
        }
        return TCL_OK;
      }

      if( (objc%2) != 0 ){
        Tcl_AppendResult(interp, "Must be an even number of arguments", 0);
        return TCL_ERROR;
      }

      /* Do a prepass checking for errors first.
      */
      for( int i=2; i < objc; i+=2 ){
        char *zRow = Tcl_GetStringFromObj(objv[i], 0);
        int row;
        if( sscanf(zRow, "%d", &row)!=1 || row < 1 ){
          Tcl_AppendResult(interp, "Bad row value ", zRow, 0);
          return TCL_ERROR;
        }
        char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);
        int height;
        if( strcmp(zValue, "default")!=0 && sscanf(zValue, "%d", &height)!=1 ){
          Tcl_AppendResult(interp, "Bad row height", 0);
          return TCL_ERROR;
        }
      }

      p->ConfigureHeights(objc-2, objv+2);
      Tcl_ResetResult(interp);
      return TCL_OK;
    }

    /*
    ** synopsis:  SS hide toggle COL ?COL...?
    **            SS hide get
    ** See HideCmd() for the details of what this command does.
    */
    case SS_HIDE:
      return p->HideCmd(objc, objv);

    /*
    ** synopsis:  SS icursor ?POS?
    ** With no arguments, returns the index of the insertion cursor within
    ** the cell that currently has the input focus.  If POS is specified,
    ** it sets the cursor position to POS.
    */
    case SS_ICURSOR:
      if( objc<2 || objc>3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?POS?");
        return TCL_ERROR;
      }
      if( objc==2 ){
        char zIc[33];
        int ic = p->fastcan.Index("active", "insert");
        sprintf(zIc, "%d", ic);
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, zIc, 0);
      } else {
        int pos = atoi(Tcl_GetStringFromObj(objv[2], 0));
        p->SetIcursor(pos);
        Tcl_ResetResult(interp);
      }
      break;

    /*
    ** synopsis: SS index INDEX ?row|col?
    ** If "row" is specified, returns just the row at INDEX.  
    ** If "col" is specified, returns just the column.
    ** If neither are specified, returns the cell at INDEX.
    */
    case SS_INDEX:
    {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "INDEX ?row|col?");
        return TCL_ERROR;
      }
      char *zRowCol = 0;
      if( objc==4 ){
        zRowCol = Tcl_GetStringFromObj(objv[3], 0);
        if( strcmp(zRowCol, "row")!=0 && strcmp(zRowCol, "col")!=0 ){
          Tcl_AppendResult(interp, "\"row\" or \"col\" must be specified", 0);
          return TCL_ERROR;
        }
      }
      char *zIndex = Tcl_GetStringFromObj(objv[2], 0);
      SsString *pCellComponent = p->GetCellComponent(zIndex, zRowCol);
      Tcl_ResetResult(interp);
      if( pCellComponent ){
        Tcl_AppendResult(interp, pCellComponent->render(), 0);
        delete pCellComponent;
      }
      break;
    }

    /*
    ** synopsis:  SS insert STRING
    ** Insert STRING at the current insertion point within the cell that
    ** currently has the input focus.
    */
    case SS_INSERT:
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "STRING");
        return TCL_ERROR;
      }
      p->InsertActive(Tcl_GetStringFromObj(objv[2], 0));
      Tcl_ResetResult(interp);
      break;

    /*
    ** synopsis:  SS minheight ROW
    ** Returns the minimum height in pixels of the given ROW.
    */
    case SS_MINHEIGHT:
    {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "ROW");
        return TCL_ERROR;
      }
      int *pWidths = new int[p->sConfigure.cols];
      for( int col = 0; col<p->sConfigure.cols; ++col ){
        pWidths[col] = p->GetColumnWidth(col);
      }
      int height = 
          p->GetRowTextHeight(atoi(Tcl_GetStringFromObj(objv[2], 0)), pWidths);
      delete [] pWidths;
      Tcl_ResetResult(interp);
      SsString res(height);
      Tcl_AppendResult(interp, res.render(), 0);
      break;
    }

    /*
    ** synopsis:  SS minwidth COL ?VALUE?
    ** If VALUE is not specified, returns the minimum width of the given COL.
    ** If VALUE is specified, sets the minimum width of COL to VALUE.
    ** But in no case can the minimum fall below the configured
    ** "-mincolwidth" value.
    */
    case SS_MINWIDTH:
    {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "COL ?VALUE?");
        return TCL_ERROR;
      }
      int col = atoi(Tcl_GetStringFromObj(objv[2], 0));
      if( objc==3 ){
        char zWidth[33];
        sprintf(zWidth, "%d", p->GetMinimumWidth(col));
        Tcl_AppendResult(interp, zWidth, 0);
      } else {
        p->SetMinimumWidth(col, atoi(Tcl_GetStringFromObj(objv[3], 0)));
      }
      break;
    }

    /*
    ** synopsis:  SS nextemptyrow ROW
    ** This will return the next row in the spreadsheet past ROW that
    ** contains no data.
    */
    case SS_NEXTEMPTYROW:
    {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "ROW");
        return TCL_ERROR;
      }
      int row = atoi(Tcl_GetStringFromObj(objv[2], 0));
      SsString res(p->GetNextRowWithoutData(row));
      Tcl_AppendResult(interp, res.render(), 0);
      break;
    }

    /*
    ** synopsis: SS redisplay ?-now? ?-trim?
    ** If "-now" is specified, the spreadsheet should be redisplayed now,
    ** rather than at idle time.
    ** If "-trim" is set, the spreadsheet's number of rows should be 
    ** trimmed if possible.
    */
    case SS_REDISPLAY:
    {
      if( objc<2 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?-now? ?-trim?");
        return TCL_ERROR;
      }

      int bNow = 0;
      int bRecalc = 0;
      for( int iobjc=2; iobjc<objc; iobjc++ ){
        char *zArg=Tcl_GetStringFromObj(objv[iobjc], 0);
        if( strcmp(zArg, "-now")==0 ){
          bNow = 1;
        } else if( strcmp(zArg, "-trim")==0 ){
          bRecalc = 1;
        } else {
          return TCL_ERROR;
        }
      }
      p->Display(bNow, bRecalc);
      break;
    }

    /*
    ** synopsis: SS see INDEX
    ** Scrolls the view so that the cell at INDEX is visible on the screen.
    */
    case SS_SEE:
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "INDEX");
        return TCL_ERROR;
      }
      p->See(Tcl_GetStringFromObj(objv[2], 0));
      Tcl_ResetResult(interp);
      break;

    /*
    ** synopsis:  SS selection anchor INDEX
    **            SS selection clear FIRSTINDEX ?LASTINDEX?
    **            SS selection includes INDEX
    **            SS selection present
    **            SS selection set ?-only? FIRSTINDEX ?LASTINDEX?
    **            SS selection setlist CELLS
    ** See SelectionCmd() for details.
    */            
    case SS_SELECTION:
      return p->SelectionCmd(objc, objv);

    /*
    ** synopsis: SS selrect bbox
    **           SS selrect draw TOPLEFTCELL BOTTOMRIGHTCELL COLOR TAG
    ** See SelrectCmd() for details.
    */
    case SS_SELRECT:
      return p->SelrectCmd(objc, objv);

    /*
    ** synopsis:  SS set INDEX ?VALUE?
    ** Sets the cell at INDEX so that it displays the given VALUE.
    ** If no VALUE is specified, returns the value of the cell at INDEX.
    */
    case SS_SET:
    {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "INDEX ?VALUE?");
        return TCL_ERROR;
      }
      char *zIndex = Tcl_GetStringFromObj(objv[2], 0);
      SsCell *pCell = p->Index(zIndex);
      if( !pCell ){
        Tcl_ResetResult(interp);
        return TCL_OK;
      }
      if( objc==3 ){
        SsString *pValue = p->fastcan.Itemcget(pCell->render(), "-text");
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, pValue->render(), 0);
        delete pValue;
        delete pCell;
      } else {
        p->Set(pCell->render(), Tcl_GetStringFromObj(objv[3], 0));
        delete pCell;
      }
      break;
    }

    /*
    ** synopsis: SS textselection clear
    **           SS textselection set INDEX FROM TO
    **           SS textselection get
    **
    ** If "clear", clears the selected intracell (text) selection.
    ** If "set INDEX FROM TO", sets the intracell text selection to
    ** be from the FROM-th character to the TO-th character.
    ** If "get", returns the range of characters that are selected.
    */
    case SS_TEXTSELECTION:
    {
      if( objc<3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "clear|set INDEX FROM TO|get");
        return TCL_ERROR;
      }
      char *zCmd = Tcl_GetStringFromObj(objv[2], 0);
      if( *zCmd=='c' && strcmp(zCmd, "clear")==0 ){
        if( objc!=3 ){
          Tcl_WrongNumArgs(interp, 3, objv, "");
          return TCL_ERROR;
        }
        p->ClearTextSelection();
        Tcl_ResetResult(interp);
      } else if(*zCmd=='s' && strcmp(zCmd, "set")==0 ){
        if( objc!=6 ){
          Tcl_WrongNumArgs(interp, 3, objv, "INDEX FROM TO");
          return TCL_ERROR;
        }
        char *zIndex = Tcl_GetStringFromObj(objv[3], 0);
        int from, to;
        if( sscanf(Tcl_GetStringFromObj(objv[4], 0), "%d", &from)!=1 ){
          Tcl_AppendResult(interp, "Bad FROM parameter", 0);
          return TCL_ERROR;
        }
        if( sscanf(Tcl_GetStringFromObj(objv[5], 0), "%d", &to)!=1 ){
          Tcl_AppendResult(interp, "Bad TO parameter", 0);
          return TCL_ERROR;
        }
        p->SetTextSelection(zIndex, from, to);
        Tcl_ResetResult(interp);
      } else if(*zCmd=='g' && strcmp(zCmd, "get")==0 ){
        if( objc!=3 ){
          Tcl_WrongNumArgs(interp, 3, objv, "");
          return TCL_ERROR;
        }
        char *zSelection = p->GetTextSelection();
        Tcl_ResetResult(interp);
        if( zSelection ){
          Tcl_AppendResult(interp, zSelection, 0);
          delete [] zSelection;
        }
      } else {
        Tcl_AppendResult(interp, "Bad textselection command ", zCmd, 0);
        return TCL_ERROR;
      }
      break;
    }

    /*
    ** synopsis: SS trim
    **
    ** This will cause rows to be trimmed from the bottom of the spreadsheet,
    ** if they contain no data and don't cause the spreadsheet to fall under
    ** a minimum number of rows.
    */
    case SS_TRIM:
    {
      if( objc!=2 ){
        Tcl_WrongNumArgs(interp, 2, objv, "");
        return TCL_ERROR;
      }
      p->TrimRows();
      break;
    }

    /*
    ** synopsis:  SS viewport
    **
    ** This will return the width and height, in pixels,
    ** of the spreadsheet viewport as a 2-valued list.
    */
    case SS_VIEWPORT:
    {
      if( objc!=2 ){
        Tcl_WrongNumArgs(interp, 2, objv, "");
        return TCL_ERROR;
      }
      SsTclList *pViewport = p->GetViewport();
      if( pViewport ){
        Tcl_SetObjResult(interp, pViewport->get_listrep());
        delete pViewport;
      }
      break;
    }

    /*
    ** synopsis:  SS width ?COL? ?VALUE? ?COL VALUE?...
    **            SS width default
    ** With no arguments, returns a list of columns and their widths in pixels.
    ** With one COL argument, returns its width in pixels.
    ** If at least one COL VALUE pair is specified, it sets each 
    ** specified COL's width to the specified VALUE.
    */
    case SS_WIDTH:
    {
      if( objc<2 ){
        Tcl_WrongNumArgs(interp, 2, objv, 
            "width default\nwidth ?COL? ?VALUE? ?COL VALUE?...");
        return TCL_ERROR;
      }

      if( objc==2 ){

        /* Return the widths of all configured columns as an even-numbered
        ** list of columns and widths
        */
        SsTclList *pWidths = p->GetAllWidths();
        Tcl_SetObjResult(interp, pWidths->get_listrep());
        delete pWidths;
        return TCL_OK;
      }

      if( objc==3 ){
        char *zArg = Tcl_GetStringFromObj(objv[2], 0);
        if( strcmp(zArg, "default")==0 ){  // reset all widths to their defaults
          p->ResetConfiguredWidths();
        } else {
          int col;
          if( sscanf(zArg, "%d", &col)!=1 ){
            Tcl_AppendResult(interp, "Argument ", zArg, " not an integer", 0);
            return TCL_ERROR;
          }
          SsString res(p->GetConfiguredWidth(col));
          Tcl_ResetResult(interp);
          Tcl_AppendResult(interp, res.render(), 0);
        }
        return TCL_OK;
      }

      if( (objc%2) != 0 ){
        Tcl_AppendResult(interp, "Must be an even number of arguments", 0);
        return TCL_ERROR;
      }

      /* Do a prepass checking for errors first.
      */
      for( int i=2; i < objc; i+=2 ){
        char *zCol = Tcl_GetStringFromObj(objv[i], 0);
        int col;
        if( sscanf(zCol, "%d", &col)!=1 || col < 0 ){
          Tcl_AppendResult(interp, "Bad column value ", zCol, 0);
          return TCL_ERROR;
        }
        char *zValue = Tcl_GetStringFromObj(objv[i+1], 0);
        int width;
        if( strcmp(zValue, "default")!=0 && sscanf(zValue, "%d", &width)!=1 ){
          Tcl_AppendResult(interp, "Bad column width ", zValue, 0);
          return TCL_ERROR;
        }
      }

      p->ConfigureWidths(objc-2, objv+2);
      Tcl_ResetResult(interp);
      return TCL_OK;
    }

    /*
    ** synopsis:  SS xview
    **            SS xview LEFTCOL
    **            SS xview moveto FRACTION
    **            SS xview scroll NUMBER WHAT
    ** With no arguments, returns a list of two values: the fractional 
    ** percentage of the entire screen width that represents the leftmost 
    ** part of the screen, and the fractional percentage of the entire
    ** screen width representing the rightmost part of the screen.
    **
    ** With one argument, scrolls the screen so that LEFTCOL is the 
    ** (non-title) column at the far left of the screen.
    **
    ** For "moveto FRACTION", makes it so that FRACTION is the fractional
    ** part of the screen that is offscreen to the left (rounding to the
    ** nearest spreadsheet non-title column boundary).
    **
    ** For "scroll NUMBER WHAT", horizontally scrolls the page 
    ** NUMBER pages or units.  WHAT is either "pages" or "units".
    */
    case SS_XVIEW:
    {
      if( objc==2 ){
        p->fastcan.Eval("xview");
      } else {
        char *zCmd = Tcl_GetStringFromObj(objv[2], 0);
        if( *zCmd=='m' && strcmp(zCmd, "moveto")==0 ){
          if( objc!=4 ){
            Tcl_WrongNumArgs(interp, 3, objv, "FRACTION");
            return TCL_ERROR;
          }
          char *zFract = Tcl_GetStringFromObj(objv[3], 0);
          float fract;
          if( sscanf(zFract, "%f", &fract)!=1 ){
            Tcl_AppendResult(interp, "Bad fraction ", zFract, 0);
            return TCL_ERROR;
          }
          p->XviewMoveto(fract);
          Tcl_ResetResult(interp);
        } else if(*zCmd=='s' && strcmp(zCmd, "scroll")==0 ){
          if( objc!=5 ){
            Tcl_WrongNumArgs(interp, 3, objv, "NUMBER WHAT");
            return TCL_ERROR;
          }
          char *zNumber = Tcl_GetStringFromObj(objv[3], 0);
          int iNum;
          if( sscanf(zNumber, "%d", &iNum)!=1 ){
            Tcl_AppendResult(interp, "Bad NUMBER argument ", zNumber, 0);
            return TCL_ERROR;
          }
          char *zWhat = Tcl_GetStringFromObj(objv[4], 0);
          if( strcmp(zWhat, "units")!=0 && strcmp(zWhat, "pages")!=0 ){
            Tcl_AppendResult(interp, "Bad WHAT argument ", zWhat, 0);
            return TCL_ERROR;
          }
          p->XviewScroll(iNum, strcmp(zWhat, "units")==0);
          Tcl_ResetResult(interp);
        } else {
          int leftcol;
          if( sscanf(zCmd, "%d", &leftcol)!=1 ){
            Tcl_AppendResult(interp, "Bad xview argument ", zCmd, 0);
            return TCL_ERROR;
          }
          p->Xscroll(++leftcol);
          Tcl_ResetResult(interp);
        }
      }
      break;
    }

    /*
    ** synopsis:  SS yview
    **            SS yview TOPROW
    **            SS yview moveto FRACTION
    **            SS yview scroll NUMBER WHAT
    **
    ** With no arguments, returns a list of two values: the fractional 
    ** percentage of the entire screen width that represents the uppermost 
    ** part of the screen, and the fractional percentage of the entire
    ** screen width representing the lowest part of the screen.
    **
    ** With one argument, scrolls the screen so that TOPROW is the (non-title)
    ** row at the top of the screen.
    **
    ** For "moveto FRACTION", makes it so that FRACTION is the fractional
    ** part of the screen that is offscreen above the screen (rounding to the
    ** nearest spreadsheet non-title row boundary).
    **
    ** For "scroll NUMBER WHAT", vertically scrolls the page 
    ** NUMBER pages or units.  WHAT is either "pages" or "units".
    */
    case SS_YVIEW:
    {
      if( objc==2 ){
        p->fastcan.Eval("yview");
      } else {
        char *zCmd = Tcl_GetStringFromObj(objv[2], 0);
        if( *zCmd=='m' && strcmp(zCmd, "moveto")==0 ){
          if( objc!=4 ){
            Tcl_WrongNumArgs(interp, 3, objv, "FRACTION");
            return TCL_ERROR;
          }
          char *zFract = Tcl_GetStringFromObj(objv[3], 0);
          float fract;
          if( sscanf(zFract, "%f", &fract)!=1 ){
            Tcl_AppendResult(interp, "Bad fraction ", zFract, 0);
            return TCL_ERROR;
          }
          p->YviewMoveto(fract);
          Tcl_ResetResult(interp);
        } else if(*zCmd=='s' && strcmp(zCmd, "scroll")==0 ){
          if( objc!=5 ){
            Tcl_WrongNumArgs(interp, 3, objv, "NUMBER WHAT");
            return TCL_ERROR;
          }
          char *zNumber = Tcl_GetStringFromObj(objv[3], 0);
          int iNum;
          if( sscanf(zNumber, "%d", &iNum)!=1 ){
            Tcl_AppendResult(interp, "Bad NUMBER argument ", zNumber);
            return TCL_ERROR;
          }
          char *zWhat = Tcl_GetStringFromObj(objv[4], 0);
          if( strcmp(zWhat, "units")!=0 && strcmp(zWhat, "pages")!=0 ){
            Tcl_AppendResult(interp, "Bad WHAT argument ", zWhat, 0);
            return TCL_ERROR;
          }
          p->YviewScroll(iNum, strcmp(zWhat, "units")==0);
          Tcl_ResetResult(interp);
        } else {
          int toprow;
          if( sscanf(zCmd, "%d", &toprow)!=1 ){
            Tcl_AppendResult(interp, "Bad yview argument ", zCmd, 0);
            return TCL_ERROR;
          }
          p->Yscroll(++toprow, 1);
          Tcl_ResetResult(interp);
        }
      }
      break;
    }

    default:
      Tcl_AppendResult(interp, "This method has not been implemented", 0);
      return TCL_ERROR;
  }

  return TCL_OK;
}

/*
** Execute the "fast_canvas" command given by the argument list.
** Return TCL_OK on success or TCL_ERROR on error.
** Return the result of the command as a Tcl result.
*/
PUBLIC int ss::FastCanvasCmd(int objc, Tcl_Obj *const* objv){
  static const char *FC_strs[] = {
    "bbox",     "find",         "gettags",  
    "index",    "itemcget",     "select",
    0,
  };
  enum FC_enum {
    FC_BBOX,    FC_FIND,        FC_GETTAGS,
    FC_INDEX,   FC_ITEMCGET,    FC_SELECT
  };

  if( objc<3 ){
    Tcl_WrongNumArgs(interp, 1, objv, "METHOD ?ARG ...?");
    return TCL_ERROR;
  }

  int index;
  if( Tcl_GetIndexFromObj(interp, objv[2], FC_strs, "option", 0, &index)){
    return TCL_ERROR;
  }

  switch( (enum FC_enum) index ){

    /*
    ** synopsis:  SS fast_canvas bbox TAG
    */
    case FC_BBOX:
    {
      if( objc!=4 ){
        Tcl_WrongNumArgs(interp, 3, objv, "TAG");
        return TCL_ERROR;
      }
      SsSetOfString *pTagSet = new SsSetOfString();
      pTagSet->insert(Tcl_GetStringFromObj(objv[3], 0));
      SsBbox *pBbox = fastcan.Bbox(pTagSet);
      delete pTagSet;
      Tcl_ResetResult(interp);
      if( pBbox ){
        Tcl_AppendResult(interp, pBbox->render(), 0);
        delete pBbox;
      }
      break;
    }

    /*
    ** synopsis:  SS fast_canvas find TAG
    */
    case FC_FIND:
    {
      if( objc!=4 ){
        Tcl_WrongNumArgs(interp, 3, objv, "TAG");
        return TCL_ERROR;
      }
      char *zTag = Tcl_GetStringFromObj(objv[3], 0);
      SsTclList *pList = new SsTclList(interp);
      for( int id=fastcan.FindFirst(zTag); id!=0; id = fastcan.FindNext(zTag) ){
        pList->append(id);
      }
      Tcl_SetObjResult(interp, pList->get_listrep());
      delete pList;
      break;
    }

    /*
    ** synopsis:  SS fast_canvas gettags CANVAS_ID
    */
    case FC_GETTAGS:
    {
      if( objc!=4 ){
        Tcl_WrongNumArgs(interp, 3, objv, "CANVAS_ID");
        return TCL_ERROR;
      }
      int id = atoi(Tcl_GetStringFromObj(objv[3], 0));
      SsTclList *pList = new SsTclList(interp);
      for( char *zTag=fastcan.IdToFirstTag(id); zTag; 
          zTag=fastcan.IdToNextTag() ){
        pList->append(zTag);
      }
      Tcl_SetObjResult(interp, pList->get_listrep());
      delete pList;
      break;
    }

    /*
    ** synopsis:  SS fast_canvas index TAG INDEX
    */
    case FC_INDEX:
    {
      if( objc!=5 ){
        Tcl_WrongNumArgs(interp, 3, objv, "TAG INDEX");
        return TCL_ERROR;
      }
      int idx = fastcan.Index(Tcl_GetStringFromObj(objv[3], 0),
          Tcl_GetStringFromObj(objv[4], 0));
      if( idx<0 ){
        Tcl_AppendResult(interp, "Could not find index", 0);
        return TCL_ERROR;
      }
      Tcl_ResetResult(interp);
      char zInt[33];
      sprintf(zInt, "%d", idx);
      Tcl_AppendResult(interp, zInt, 0);
      break;
    }

    /*
    ** synopsis:  SS fast_canvas itemcget TAG OPTION
    */
    case FC_ITEMCGET:
    {
      if( objc!=5 ){
        Tcl_WrongNumArgs(interp, 3, objv, "TAG OPTION");
        return TCL_ERROR;
      }
      SsString *pResult = fastcan.Itemcget(Tcl_GetStringFromObj(objv[3], 0),
          Tcl_GetStringFromObj(objv[4], 0));
      Tcl_ResetResult(interp);
      if( pResult ){
        Tcl_AppendResult(interp, pResult->render(), 0);
        delete pResult;
      }
      break;
    }

    /*
    ** synopsis:  SS fast_canvas select from TAG INDEX
    **            SS fast_canvas select to TAG INDEX
    **            SS fast_canvas select adjust TAG INDEX
    */
    case FC_SELECT:
    {
      if( objc!=6 ){
        Tcl_WrongNumArgs(interp, 3, objv, "from|to|adjust TAG INDEX");
        return TCL_ERROR;
      }
      char *zSelType = Tcl_GetStringFromObj(objv[3], 0);
      char *zTag = Tcl_GetStringFromObj(objv[4], 0);
      char *zIndex = Tcl_GetStringFromObj(objv[5], 0);
      if( *zSelType=='f' && strcmp(zSelType, "from")==0 ){
        fastcan.SelectFrom(zTag, zIndex);
      } else if( *zSelType=='t' && strcmp(zSelType, "to")==0 ){
        fastcan.SelectTo(zTag, zIndex);
      } else if( *zSelType=='a' && strcmp(zSelType, "adjust")==0 ){
        fastcan.SelectAdjust(zTag, zIndex);
      } else {
        return TCL_ERROR;
      }
      break;
    }
  }
  return TCL_OK;
}

/*
** Execute the "hide" command given by the argument list.
** Return TCL_OK on success or TCL_ERROR on error.
** Return the result of the command as a Tcl result.
*/
PUBLIC int ss::HideCmd(int objc, Tcl_Obj *const* objv){
  static const char *HIDE_strs[] = {
    "get",     "toggle",
    0,
  };
  enum HIDE_enum {
    HIDE_GET,  HIDE_TOGGLE,
  };
  if( objc<3 ){
    Tcl_WrongNumArgs(interp, 1, objv, "METHOD ?ARG ...?");
    return TCL_ERROR;
  }

  int index;
  if( Tcl_GetIndexFromObj(interp, objv[2], HIDE_strs, "option", 0, &index)){
    return TCL_ERROR;
  }

  switch( (enum HIDE_enum) index ){

    /*
    ** synopsis:  SS hide get ?COL?
    ** Return the list of currently hidden columns.
    ** Or if COL is specified, return 1 if it's hidden and 0 otherwise.
    */
    case HIDE_GET:
      if( objc!=3 && objc!=4 ){
        Tcl_WrongNumArgs(interp, 3, objv, "?COL?");
        return TCL_ERROR;
      }
      if( objc==3 ){
        SsTclList result(interp);
        for( int col=0; col<sConfigure.cols; ++col ){
          if( GetHidden(col) ){
            result.append(col);
          }
        }
        Tcl_SetObjResult(interp, result.get_listrep());
      } else {
        int col = -1;
        Tcl_GetIntFromObj(0, objv[3], &col);
        if( col>=0 && col<sConfigure.cols ){
          Tcl_AppendResult(interp, GetHidden(col)? "1" : "0", 0);
        } else {
          Tcl_AppendResult(interp, "Bad column ", 
              Tcl_GetStringFromObj(objv[3], 0), 0);
          return TCL_ERROR;
        }
      }
      break;

    /*
    ** synopsis: HIDE toggle COL ?COL...?
    ** Toggle the "hidden" state for the given list of columns.
    */
    case HIDE_TOGGLE: {
      if( objc<4 ){
        Tcl_WrongNumArgs(interp, 3, objv, "COL ?COL...?");
        return TCL_ERROR;
      }
      for( int i=3; i!=objc; ++i ){
        int col = -1;
        Tcl_GetIntFromObj(0, objv[i], &col);
        if( col<0 || col>=sConfigure.cols ){
          Tcl_AppendResult(interp, "Bad column ", 
              Tcl_GetStringFromObj(objv[i], 0), 0);
          return TCL_ERROR;
        }
      }
      for( int i=3; i!=objc; ++i ){
        int col = -1;
        Tcl_GetIntFromObj(0, objv[i], &col);
        ToggleHide(col);
      }
      break;
    }
  }
  return TCL_OK;
}

/*
** Execute the "selection" command given by the argument list.
** Return TCL_OK on success or TCL_ERROR on error.
** Return the result of the command as a Tcl result.
*/
PUBLIC int ss::SelectionCmd(int objc, Tcl_Obj *const* objv){
  static const char *SEL_strs[] = {
    "anchor",     "clear",         "cleardata",
    "includes",   "present",       "set",
    "setlist",
    0,
  };
  enum SEL_enum {
    SEL_ANCHOR,    SEL_CLEAR,        SEL_CLEARDATA,
    SEL_INCLUDES,  SEL_PRESENT,      SEL_SET,
    SEL_SETLIST,
  };

  if( objc<3 ){
    Tcl_WrongNumArgs(interp, 1, objv, "METHOD ?ARG ...?");
    return TCL_ERROR;
  }

  int index;
  if( Tcl_GetIndexFromObj(interp, objv[2], SEL_strs, "option", 0, &index)){
    return TCL_ERROR;
  }

  switch( (enum SEL_enum) index ){

    /*
    ** synopsis:  SS selection anchor INDEX
    ** Begin a selection by placing the "selection anchor" at the cell
    ** referenced by INDEX.
    */
    case SEL_ANCHOR:
    {
      if( objc!=4 ){
        Tcl_WrongNumArgs(interp, 3, objv, "INDEX");
        return TCL_ERROR;
      }
      char *zIndex = Tcl_GetStringFromObj(objv[3], 0);
      SsCell *pCell = Index(zIndex);
      if( !pCell ){
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "No such index ", zIndex, 0);
        return TCL_ERROR;
      }

      SetSelectionAnchor(pCell->render());
      delete pCell;
      Tcl_ResetResult(interp);
      break;
    }

    /*
    ** synopsis:  SS selection clear FIRSTINDEX ?LASTINDEX?
    ** Deselect all cells from FIRSTINDEX to LASTINDEX inclusive.
    ** If LASTINDEX isn't specified, just deselect FIRSTINDEX.
    */
    case SEL_CLEAR:
    {
      if( objc<4 ||objc>5 ){
        Tcl_WrongNumArgs(interp, 3, objv, "FIRSTINDEX ?LASTINDEX?");
        return TCL_ERROR;
      }
      char *zFirstIndex = Tcl_GetStringFromObj(objv[3], 0);
      char *zLastIndex = zFirstIndex;
      if( objc==5 ){
        zLastIndex = Tcl_GetStringFromObj(objv[4], 0);
      }
      SelectionClearCells(zFirstIndex, zLastIndex);
      Tcl_ResetResult(interp);
      break;
    }

    /*
    ** synopsis: SS selection cleardata
    ** Clears all of the data from the selected cells.
    */
    case SEL_CLEARDATA:
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 3, objv, "");
        return TCL_ERROR;
      }
      SelectionClearData();
      Tcl_ResetResult(interp);
      break;

    /*
    ** synopsis:  SS selection includes INDEX
    ** Return 1 if the selection currently includes the cell specified by
    ** INDEX, or 0 otherwise.
    */
    case SEL_INCLUDES:
    {
      if( objc!=4 ){
        Tcl_WrongNumArgs(interp, 3, objv, "INDEX");
        return TCL_ERROR;
      }
      int bIncludes = SelectionIncludes(Tcl_GetStringFromObj(objv[3], 0));
      char zInt[33];
      sprintf(zInt, "%d", bIncludes);
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp, zInt, 0);
      break;
    }
    
    /*
    ** synopsis:  SS selection present
    ** Return 1 if a selection exists, or 0 if it does not.
    */
    case SEL_PRESENT:
    {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 3, objv, "");
        return TCL_ERROR;
      }
      int bPresent = SelectionPresent();
      char zInt[33];
      sprintf(zInt, "%d", bPresent);
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp, zInt, 0);
      break;
    }

    /*
    ** synopsis:  SS selection set ?-only? FIRSTINDEX ?LASTINDEX?
    ** If "-only" is specified, then the selection is set to just the
    ** cells between FIRSTINDEX and LASTINDEX inclusive; otherwise,
    ** the new cells are added to the current selection.
    */
    case SEL_SET:
    {
      if( objc<4 ||objc>6 ){
        Tcl_WrongNumArgs(interp, 3, objv, "?-only? FIRSTINDEX ?LASTINDEX?");
        return TCL_ERROR;
      }
      int iArg = 3;
      int bOnly = 0;
      char *zFirstToken = Tcl_GetStringFromObj(objv[iArg], 0);
      if( *zFirstToken=='-' && strcmp(zFirstToken, "-only")==0 ){
        if( objc==4 ){
          Tcl_WrongNumArgs(interp, 3, objv, "?-only? FIRSTINDEX ?LASTINDEX?");
          return TCL_ERROR;
        }
        bOnly++;
        iArg++;
      }
      char *zFirstIndex = Tcl_GetStringFromObj(objv[iArg++], 0);
      char *zLastIndex = zFirstIndex;
      if( objc==iArg+1 ){
        zLastIndex = Tcl_GetStringFromObj(objv[iArg], 0);
      }
      SelectionSet(bOnly, zFirstIndex, zLastIndex);
      Tcl_ResetResult(interp);
      break;
    }

    /*
    ** synopsis:  SS selection setlist CELLS
    ** Set the selection to be the given list of CELLS.
    */
    case SEL_SETLIST:
    {
      if( objc!=4 ){
        Tcl_WrongNumArgs(interp, 3, objv, "CELLS");
        return TCL_ERROR;
      }
      Tcl_Obj *pList = objv[3];
      Tcl_Obj **ppElem;
      int nElts;
      if( Tcl_ListObjGetElements(interp, pList, &nElts, &ppElem)!=TCL_OK ){
        return TCL_ERROR;
      }
      SelectionSetlist(nElts, ppElem);
      break;
    }
  }
  return TCL_OK;
}
/*
** Execute the "selrect" command given by the argument list.
** Return TCL_OK on success or TCL_ERROR on error.
** Return the result of the command as a Tcl result.
*/
PUBLIC int ss::SelrectCmd(int objc, Tcl_Obj *const* objv){
  static const char *SELRECT_strs[] = {
    "bbox",         "draw",
    0,
  };
  enum SELRECT_enum {
    SELRECT_BBOX,   SELRECT_DRAW,
  };

  if( objc<3 ){
    Tcl_WrongNumArgs(interp, 1, objv, "METHOD ?ARG ...?");
    return TCL_ERROR;
  }

  int index;
  if( Tcl_GetIndexFromObj(interp, objv[2], SELRECT_strs, "option", 0, &index)){
    return TCL_ERROR;
  }

  switch( (enum SELRECT_enum) index ){

    /*
    ** synopsis:  SS selrect bbox
    ** Return the "bounding box" of the selection rectangle in the form
    ** "row0 col0 row1 col1".  The upperleft cell in the selection
    ** rectangle is at row0,col0 and the lowerright cell is at row1,col1.
    */
    case SELRECT_BBOX:
    {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 3, objv, "");
        return TCL_ERROR;
      }
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp, GetSelrectBbox(), 0);
      break;
    }

    /*
    ** synopsis:  SS selrect draw TOPLEFTCELL BOTTOMRIGHTCELL COLOR TAG
    ** Draw the selection rectangle so that TOPLEFTCELL and BOTTOMRIGHTCELL
    ** define its bounding box.  Color the rectangle boundary with COLOR.
    ** The canvas will tag the rectangle glyph with the given TAG.
    */
    case SELRECT_DRAW:
    {
      if( objc!=7 ){
        Tcl_WrongNumArgs(interp, 3, objv, "TOPLEFTCELL BOTTOMRIGHTCELL"
                         " COLOR TAG");
        return TCL_ERROR;
      }
      SsCell tlcell(Tcl_GetStringFromObj(objv[3], 0));
      SsCell brcell(Tcl_GetStringFromObj(objv[4], 0));
      char *zColor = Tcl_GetStringFromObj(objv[5], 0);
      char *zTag = Tcl_GetStringFromObj(objv[6], 0);
      int bOk = DrawRectangle(tlcell, brcell, 2, zColor, zTag);
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp, bOk ? "1" : "0", 0);
      break;
    }
  }
  return TCL_OK;
}

/*
** This is a Tcl command used to create a new ss object.
*/
PUBLIC STATIC int ss::CreateProc(
  void *NotUsed,
  Tcl_Interp *interp,
  int argc,
  const char **argv
){
  if( argc!=2 ){
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
       " WIDGET-NAME\"", 0);
    return TCL_ERROR;
  }
  ss *p = new ss(interp, argv[1]);
  Tcl_CreateObjCommand(interp, argv[1], ss::CmdProc, (void*)p, ss::DeleteProc);
  Tcl_AppendResult(interp, argv[1], 0);
  return TCL_OK;
}

/*
** Initialize the ss command
*/
extern "C" int Ss_Init(Tcl_Interp *interp){
  Tcl_CreateCommand(interp, "::ss", (Tcl_CmdProc*)ss::CreateProc, 0, 0);
  Tcl_CreateCommand(interp, "::ss_display", (Tcl_CmdProc*)ss::DisplayCmd, 0, 0);
  return TCL_OK;
}
